# import hydra 
# from omegaconf import OmegaConf

# from move.conf.schema import MOVEConfig

# @hydra.main(config_path="conf", config_name="main")
# def main(config: MOVEConfig) -> str:
#     print(config.model.continuous_weights)
#     print(OmegaConf.to_yaml(config))
    
if __name__ == "__main__":
    print('hi')
    pass
#     main()




import os
import torch
import numpy as np
from torch.utils.data import DataLoader
import umap.umap_ as umap

import matplotlib.pyplot as plt
plt.style.use('seaborn-whitegrid')

from scipy.stats.stats import pearsonr
from sklearn.cluster import KMeans
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.metrics import adjusted_rand_score
import pandas as pd
import seaborn as sns
from omegaconf import OmegaConf
import itertools 
from tqdm import tqdm
from collections import Counter, defaultdict
from scipy import stats

from move.utils.data_utils import initiate_default_dicts
from move.utils import dataloaders
from move.models import vae


def get_top10_stability(nHiddens, nLatents, drop_outs, nLayers, repeat, latents, batch_sizes, nBeta):
    # Todo: TOP10 not used description by Rosa
    npatient = list(latents.values())[0][0].shape[0]
    top10_changes, stability_top10 = initiate_default_dicts(0, 2) #TODOS: top10 changes not needed? - mention, not used
    
    stability_top10_df = [] 
        
    iters = itertools.product(nHiddens, nLatents, nLayers, drop_outs, nBeta)
    for nHidden, nLatent, nl, drop, b in iters:
        
        max_pos_values_init = list()
        old_sum_max = list()
        
        name = str([nHidden] * nl) + "+" + str(nLatent) + ", do: " + str(drop) +", b: " + str(b)

        print(name)
        top10_changes[name] = [ [] for i in range(npatient) ]
     
        for r in range(repeat):
            cos_sim = cosine_similarity(latents[name][r])
            corr = pd.DataFrame(cos_sim)
                
            step = list()
            for index, row in corr.iterrows():
                if r == 0:
                    max_pos = np.asarray(row.argsort()[11:][::-1][1:11])
                    max_pos_values_init.append(max_pos)
                    #summed_max = np.mean(row[max_pos])
                    old_sum_max.append(row[max_pos])
                else:
                    old_pos = max_pos_values_init[index]
                    old_sum = old_sum_max[index]
                    #summed_max = np.mean(row[old_pos])
                    top10_changes[name][index].append(np.mean(abs(old_sum - row[old_pos])))
                    step.append(np.mean(abs(old_sum - row[old_pos])))
            print(r)
            if r != 0:
                stability_top10[name].append(np.mean(step))
                stability_top10_df.append({
                    'num_hidden': nHidden,
                    'num_latent': nLatent,
                    'num_layers': nl,
                    'dropout': drop,
                    'beta': b,
                    'batch_sizes': batch_sizes, 
                    'repeats': r,
                    'difference': np.mean(step)
                    })
    stability_top10_df = pd.DataFrame(stability_top10_df)
    return(stability_top10, stability_top10_df)


def calculate_latent(nHiddens, nLatents, drop_outs, repeat, nLayers, nBeta, latents):

    npatient = list(latents.values())[0][0].shape[0]
    total_changes, stability_total, rand_index = initiate_default_dicts(0, 3)
     
    iters = itertools.product(nHiddens, nLatents, nLayers, drop_outs, nBeta)
    for nHidden, nLatent, nl, drop, b in iters:
       
        pos_values_init = list() 
        old_rows = list()
          
        name = str([nHidden] * nl) + "+" + str(nLatent) + ", do: " + str(drop) +", b: " + str(b)
        total_changes[name] = [ [] for i in range(npatient) ]

        for r in range(repeat):
            cos_sim = cosine_similarity(latents[name][r])

            corr = pd.DataFrame(cos_sim)
            step = list()
            for index, row in corr.iterrows():
                if r == 0:
                    max_pos = np.asarray(row.argsort()[:][::-1][1:])
                    pos_values_init.append(max_pos)
                    old_rows.append(row[max_pos])
                else:
                    old_pos = pos_values_init[index]
                    old_row = old_rows[index]
                    total_changes[name][index].append(np.mean(abs(old_row - row[old_pos])))
                    step.append(np.mean(abs(old_row - row[old_pos])))

            if r != 0:
                kmeans = KMeans(n_clusters=4)
                rand_tmp = []
                for i in range(0,100):
                    kmeans = kmeans.fit(latents[name][r])
                    labels = kmeans.predict(latents[name][r])
                    rand_tmp.append(adjusted_rand_score(true_labels, labels))

                rand_index[name].append(np.mean(rand_tmp))
                stability_total[name].append(np.mean(step))
            else:
                kmeans = KMeans(n_clusters=4)
                kmeans = kmeans.fit(latents[name][r])
                true_labels = kmeans.predict(latents[name][r])

    stability_total = pd.DataFrame(stability_total)
    return(stability_total, rand_index)


def get_latents(best_model, train_loader, kld_w=1): 
     
    # Extracting the latent space
    train_test_loader = DataLoader(dataset=train_loader.dataset, batch_size=1, 
                                             drop_last=False, shuffle=False, 
                                             pin_memory=train_loader.pin_memory) 

    latent, latent_var, cat_recon, cat_class, \
    con_recon, loss, likelihood = best_model.latent(train_test_loader, 1)

    con_recon = np.array(con_recon)
    con_recon = torch.from_numpy(con_recon)
     
    return latent, latent_var, cat_recon, cat_class, con_recon, loss, likelihood


def calc_categorical_reconstruction_acc(cat_shapes, cat_class, cat_recon):
    # Calculate the categorical reconstruction accuracy
    cat_true_recon = []
    cat_total_recon = []
    pos = 0
    for s in cat_shapes:
        n = s[1]
        cat_class_tmp = cat_class[:,pos:(n + pos)]
        cat_recon_tmp = cat_recon[:,pos:(n + pos)]

        missing_cat = cat_recon_tmp[cat_class_tmp == -1]
        diff_cat = cat_class_tmp - cat_recon_tmp

        diff_cat[diff_cat != 0] = -1
        true_cat = diff_cat[diff_cat == 0]
        false_cat = diff_cat[diff_cat != 0]

        cat_true = len(true_cat)/(float(diff_cat.size) - missing_cat.size)
        cat_true_recon.append(cat_true)
        diff_cat[diff_cat == 0] = 1
        diff_cat[diff_cat != 1] = 0
        cat_total_recon.append(np.count_nonzero(diff_cat, 1) / diff_cat.shape[1])
        pos += n
    return(cat_total_recon)


def calc_continuous_reconstruction_acc(con_shapes, con_recon, train_loader):
    # Calculate the continuous reconstruction accuracy
    total_shape = 0
    true_recon = []
    cos_values = []
    all_values = []
    for s in con_shapes:
        cor_con = list()
        cos_con = list()
        all_val = list()
        for n in range(0, con_recon.shape[0]):
            con_no_missing = train_loader.dataset.con_all[n,total_shape:(s + total_shape - 1)][train_loader.dataset.con_all[n,total_shape:(s + total_shape - 1)] != 0]
            if len(con_no_missing) <= 1:
                all_val.append(np.nan)
                continue
            con_out_no_missing = con_recon[n,total_shape:(s + total_shape - 1)][train_loader.dataset.con_all[n,total_shape:(s + total_shape - 1)] != 0]
            cor = pearsonr(con_no_missing, con_out_no_missing)[0]
            cor_con.append(cor)

            com = np.vstack([con_no_missing, con_out_no_missing])
            cos = cosine_similarity(com)[0,1]
            cos_con.append(cos)
            all_val.append(cos)

        cor_con = np.array(cor_con)
        cos_con = np.array(cos_con)
        cos_values.append(cos_con)
        all_values.append(np.array(all_val))
        true_recon.append(len(cos_con[cos_con >= 0.9]) / len(cos_con))
        total_shape += s
    return(all_values)

def get_embedding(path, latent):
    results_folder = path + 'results/'
    isExist = os.path.exists(results_folder)
    if not isExist:
         os.makedirs(results_folder)

     # UMAP
    reducer = umap.UMAP()
    embedding = reducer.fit_transform(latent)
    np.save(path + "results/embedding.npy", embedding)
    return(embedding)

def get_feature_data(data_type, feature_of_interest, cat_list, #TODO: negative values goes down to -2, while positive only until 1
                           con_list, cat_names, con_names):
     
    if data_type=='categorical':
        cat_list_integer = [np.argmax(cat, axis=-1) for cat in cat_list]
        np_data_ints = np.concatenate(cat_list_integer, axis=-1)
        headers = cat_names
    elif data_type=='continuous':
        np_data_ints = np.concatenate(con_list, axis=-1)
        headers = con_names
    else:
        raise ValueError("Wrong data type was selected")
     
    feature_data = np_data_ints[:,list(headers).index(feature_of_interest)]
     
    return(feature_data, headers)

def get_pearsonr(feature_of_interest, embedding, 
                 cat_list, con_list, cat_names, con_names):
     
    if feature_of_interest in cat_names:
        data_type = 'categorical'
    elif feature_of_interest in con_names:
        data_type = 'continuous'
    else:
        raise ValueError("feature_of_interest is not in cat_names or con_names")
    
    feature_data, _ = get_feature_data(data_type, feature_of_interest, 
                                                   cat_list, con_list, 
                                                   cat_names, con_names)
     
    # Correlate embedding with variable 
    pearson_0dim = pearsonr(embedding[:,0], feature_data)
    pearson_1dim = pearsonr(embedding[:,1], feature_data)
     
    return(pearson_0dim, pearson_1dim)

def get_feature_importance_categorical(model, train_loader, latent, kld_w=1): #Which kld_w and train_loader # should not matter equal to 1

    all_diffs = []
    sum_diffs = []
    sum_diffs_abs = []
    total_diffs = []
    loss_cat = []
    pos = 0
    cat_in = train_loader.dataset.cat_all
    cat_shapes = train_loader.dataset.cat_shapes

    for cat_shape in cat_shapes:
        cat_dataset = cat_in[:, pos:(cat_shape[1]*cat_shape[2] + pos)]
        cat_dataset = np.array(cat_dataset.view(cat_in.shape[0], cat_shape[1], cat_shape[2]))
        for feature_index in tqdm(range(cat_shape[1])):

            new_cat = np.copy(cat_dataset)
            new_cat[:,feature_index,:] = 0
            new_cat = new_cat.reshape(new_cat.shape[0], -1)
            input_cat = np.copy(cat_in)
            input_cat[:, pos:(cat_shape[1]*cat_shape[2] + pos)] = new_cat
            input_cat = torch.from_numpy(input_cat)

            dataset = dataloaders.MOVEDataset(input_cat, train_loader.dataset.con_all, 
                                                train_loader.dataset.con_shapes, 
                                                train_loader.dataset.cat_shapes)

            new_loader = DataLoader(dataset, batch_size=1, 
                                        drop_last=False, shuffle=False, 
                                        pin_memory=train_loader.pin_memory) # removed num_workers=1,

            out = model.latent(new_loader, kld_w)

            new_latent_vector = out[0]
            diff = latent-new_latent_vector
            diff_abs = np.abs(latent-new_latent_vector)
            loss_cat.append(out[-1])
            all_diffs.append(diff)
            sum_diffs.append(np.sum(diff, axis = 1))
            sum_diffs_abs.append(np.sum(diff_abs, axis = 1))
            total_diffs.append(np.sum(diff))
#             break #added

    all_diffs_cat_np = np.asarray(all_diffs)
    sum_diffs_cat_np = np.asarray(sum_diffs)
    sum_diffs_cat_abs_np = np.asarray(sum_diffs_abs)
    total_diffs_cat_np = np.asarray(total_diffs)
    return(all_diffs, all_diffs_cat_np, sum_diffs_cat_np, sum_diffs_cat_abs_np, total_diffs_cat_np)


def get_feature_importance_continuous(model, train_loader, mask, latent, kld_w=1):
# Feature importance continuous

    all_diffs_con = []
    sum_diffs_con = []
    sum_diffs_con_abs = []
    total_diffs_con = []
    loss_con = []
    con_shape = train_loader.dataset.con_all.shape[1]
    for feature_index in tqdm(range(con_shape)):

        new_con = np.array(train_loader.dataset.con_all)
        new_con[:,feature_index] = 0
        new_con = torch.from_numpy(new_con)

        dataset = dataloaders.MOVEDataset(train_loader.dataset.cat_all, new_con,
                                            train_loader.dataset.con_shapes,
                                            train_loader.dataset.cat_shapes)

        new_loader = DataLoader(dataset, batch_size=len(mask), 
                                      drop_last=False, shuffle=False, 
                                      pin_memory=train_loader.pin_memory) #removed num_workers=1,

        out = model.latent(new_loader, kld_w)

        new_latent_vector = out[0]
        loss_con.append(out[-1])
        diff_abs = np.abs(latent-new_latent_vector)
        diff = latent-new_latent_vector
        all_diffs_con.append(diff)
        sum_diffs_con.append(np.sum(diff, axis = 1))
        sum_diffs_con_abs.append(np.sum(diff_abs, axis = 1))
        total_diffs_con.append(np.sum(diff))
#         break #added

    all_diffs_con_np = np.asarray(all_diffs_con)
    sum_diffs_con_np = np.asarray(sum_diffs_con)
    sum_diffs_con_abs_np = np.asarray(sum_diffs_con_abs)
    total_diffs_con_np = np.asarray(total_diffs_con)
    return(all_diffs_con_np, sum_diffs_con_np, sum_diffs_con_abs_np, total_diffs_con_np)


def save_feat_results(path, all_diffs, sum_diffs, sum_diffs_abs, total_diffs, 
                      all_diffs_con, sum_diffs_con, sum_diffs_con_abs, total_diffs_con):
     
    # Save results
    all_diffs_both = np.concatenate((all_diffs, all_diffs_con), axis=0)
    sum_diffs_both = np.concatenate((sum_diffs, sum_diffs_con), axis=0)
    sum_diffs_both_abs = np.concatenate((sum_diffs_abs, sum_diffs_con_abs), axis=0)
    total_diffs_both = np.concatenate((total_diffs, total_diffs_con), axis=0)

    np.save(path + "results/all_SHAP_diffs.npy", all_diffs_both)
    np.save(path + "results/sum_diffs.npy", sum_diffs_both)
    np.save(path + "results/sum_diffs_abs.npy", sum_diffs_both_abs)
    np.save(path + "results/total_diffs_final.npy", total_diffs_both)

def get_feat_importance_on_weights(path, model, train_loader, cat_names, con_names):
    #Based on weights
     
    cat_shapes = train_loader.dataset.cat_shapes
    # get weights
    w = np.array(model.encoderlayers[0].weight.data.to("cpu")) 
    w_sum = np.sum(np.abs(w), axis=0)

    w_sum_con = w_sum[train_loader.dataset.cat_all.shape[1]:]
    w_sum_cat = w_sum[0:train_loader.dataset.cat_all.shape[1]]

    # Get the sum of each input for each categorical one-hot vector
    w_sum_cat_final = []
    pos = 0
    for s in cat_shapes:
        n = s[1] * s[2]
        print(n)
        cat_w_sum_tmp = w_sum_cat[pos:(n + pos)]
        cat_w_sum_tmp = cat_w_sum_tmp.reshape(s[1], s[2])
        sum_d = np.sum(cat_w_sum_tmp, axis=1)
        w_sum_cat_final.extend(sum_d)


    feature_order_cat = np.flip(np.argsort(w_sum_cat_final))
    features_w_cat = cat_names[feature_order_cat]
    # Save the 
    tmp_pd = pd.DataFrame(np.array(w_sum_cat_final)[feature_order_cat], index=features_w_cat)
    tmp_pd.T.to_csv(path + "results/importance_w_cat.txt")

    feature_order = np.flip(np.argsort(w_sum_con))
    features_w_con = con_names[feature_order]

    tmp_pd = pd.DataFrame(w_sum_con[feature_order], index=features_w_con)
    tmp_pd.T.to_csv(path + "results/importance_w_con.txt")
     
def cal_reconstruction_change(recon_results, repeats):
    recon_average = dict()
    for l in recon_results.keys():
        average = defaultdict(dict)
        for r in range(len(recon_results[l])):
            for d in range(len(recon_results[l][r])):
                tmp_recon = recon_results[l][r][d]
                if d in average:
                    average[d] = np.add(average[d], tmp_recon)
                else:
                    average[d] = tmp_recon
        a = {k: (v / repeats) for k, v in average.items()}
        recon_average[l] = a
    return(recon_average)


def overlapping_hits(nLatents, cor_results, repeats, con_names, drug): # TODOs: con_names and drug come from function return
    sig_hits = defaultdict(dict)
    overlaps_d = defaultdict(list)
    counts = list()

    new_list = nLatents[::-1]

    median_p_val = defaultdict(dict)
    for l in range(len(new_list)):
        for d in range(cor_results[0][new_list[l]].shape[0]):
            hits_tmp = list()
            p_cors = defaultdict(list)
            for repeat in range(repeats):

                ns = con_names[cor_results[repeat][new_list[l]][d,:] <= 0.05]

                p_cor = cor_results[repeat][new_list[l]][d,:]
                p_cor = p_cor[p_cor <= 0.05]
                for i,ns_t in enumerate(ns):
                    p_cors[ns_t].append(p_cor[i])

                hits_tmp.extend(ns)

            overlap_tmp = [hits_tmp.count(x) for x in np.unique(hits_tmp)]
            overlap = np.array(np.unique(hits_tmp))[np.array(overlap_tmp) >= 5]
            m_p = []
            for o_t in overlap:
                m_p.append(np.median(p_cors[o_t]))

            sig_hits[d][new_list[l]] = overlap
            median_p_val[d][new_list[l]] = m_p
    return(sig_hits, median_p_val)

def identify_high_supported_hits(sig_hits, drug_h, version, path): # drug_h comes from the function
    result = dict()
    collected_overlap = defaultdict(list)
    all_hits = list()
    for d in sig_hits:
        result[drug_h[d]] = defaultdict(list)
        all_h = []
        for l in sig_hits[d]:
            all_h.extend(sig_hits[d][l])

        for x in set(all_h):
            if all_h.count(x) >=3:
                result[drug_h[d]]['high'].append(x)
            elif all_h.count(x) == 2:
                result[drug_h[d]]['medium'].append(x)
            else:
                result[drug_h[d]]['low'].append(x)

            if all_h.count(x) >= 3:
                collected_overlap[drug_h[d]].append(x)
                all_hits.append(x)
    
    # Save result
    np.save(path + "results/results_confidence_" + version + ".npy", result)

    return(all_hits, collected_overlap)




def report_values(path, sig_hits, median_p_val, drug_h, all_hits, con_names): #TODO: drugs come from defined func

    results_folder = path + 'results/sig_ci_files'
    isExist = os.path.exists(results_folder)
    if not isExist:
        os.makedirs(results_folder)

    p_vals = list()
    for d in sig_hits:
        p_vals_col = []
        p_vals_tmp = defaultdict(list)
        for l in sig_hits[d]:
            name_s = sig_hits[d][l]
            ps = median_p_val[d][l]
            for i,ns_1 in enumerate(name_s):
                if ns_1 in collected_overlap[drug_h[d]]:
                    p_vals_tmp[ns_1].append(ps[i])

        m_p_vals_tmp = dict()
        for ns_2 in p_vals_tmp:
            m_p_vals_tmp[ns_2] = np.median(p_vals_tmp[ns_2])

        for a_h in all_hits:
            if a_h in m_p_vals_tmp:
                p_vals_col.append(m_p_vals_tmp[a_h])
            else:
                p_vals_col.append('ns')

        p_vals.append(p_vals_col)

    p_vals_df = pd.DataFrame(p_vals, index=drug_h, columns=all_hits)

    # Save the files for each continuous dataset
    for i,al_con in enumerate(con_names):
        sig_drug_names = np.intersect1d(all_hits, al_con)
        df_tmp = p_vals_df.loc[:, sig_drug_names]
        df_tmp.T.to_csv(path + "results/sig_ci_files/" + con_names[i] + "_p_vals.txt", sep = "\t")


def get_change_in_reconstruction(recon_average, groups, drug, drug_h, con_names, collected_overlap, sig_hits, con_all, version, path, types): 
#     types = [[1, 0]] #TODOs: Should types be really like this?  #Change only in this notebook

    recon_average_corr_all = dict()
    counts_average_all = dict()
    recon_average_corr_all_indi = dict()
    counts_indi = dict()

    for l in recon_average.keys():
        for d in recon_average[l].keys():
            print(d)
            tmp_recon = np.copy(recon_average[l][d])
            gr = groups[d]
            g = [not (np.all(a_s == types[0]) or (np.all(a_s == [0,0]))) for a_s in drug[gr,d,:]]
            tmp_recon = tmp_recon[g]

            if d not in recon_average_corr_all:
                recon_average_corr_all[d] = [0] * len(con_names)
                counts_average_all[d] = [0] * len(con_names)
                recon_average_corr_all_indi[d] = [[0] * tmp_recon.shape[1]] * tmp_recon.shape[0]
                counts_indi[d] = [[0] * tmp_recon.shape[1]] * tmp_recon.shape[0]

            for f in tqdm(range(tmp_recon.shape[1])):
                if con_names[f] in collected_overlap[drug_h[d]]:
                    if con_names[f] in sig_hits[d][l]:
                        tmp_vals = con_all[groups[d],f]
                        tmp_vals = tmp_vals[g]
                        avg_tmp = np.mean(tmp_recon[tmp_vals != 0,f])
                        recon_average_corr_all[d][f] += avg_tmp
                        counts_average_all[d][f] += 1

                        for indi in range(tmp_recon.shape[0]):
                            if tmp_vals[indi] != 0:
                                recon_average_corr_all_indi[d][indi][f] += tmp_recon[indi,f]
                                counts_indi[d][indi][f] += 1
                else:
                    tmp_vals = con_all[groups[d],f]
                    tmp_vals = tmp_vals[g]
                    avg_tmp = np.mean(tmp_recon[tmp_vals != 0,f])
                    recon_average_corr_all[d][f] += avg_tmp
                    counts_average_all[d][f] += 1

                    for indi in range(tmp_recon.shape[0]):
                        if tmp_vals[indi] != 0:
                            recon_average_corr_all_indi[d][indi][f] += tmp_recon[indi,f]
                            counts_indi[d][indi][f] += 1
#         break
    recon_average_corr_new_all = list()
    recon_average_corr_all_indi_new = dict()
    for d in recon_average_corr_all.keys():
        print(d)
        counts_tmp = np.array(counts_average_all[d])
        tmp_l = np.array(recon_average_corr_all[d])[counts_tmp != 0]
        included_names = con_names[counts_tmp != 0]
        counts_tmp = counts_tmp[counts_tmp != 0]
        recon_average_corr_new_all.append(tmp_l/counts_tmp)

        tmp_recon_average_corr_all_indi_new = list()
        for f in range(len(recon_average_corr_all_indi[d])):
            tmp_recon_average_corr_all_indi_new.append(np.array(recon_average_corr_all_indi[d][f]) / np.array(counts_indi[d][f]))

        recon_average_corr_all_indi_new[d] = np.transpose(np.array(tmp_recon_average_corr_all_indi_new))

    recon_average_corr_new_all = np.array(recon_average_corr_new_all)

     # Save recon results
    np.save(path + "results/results_confidence_recon_all_" + version + ".npy", recon_average_corr_new_all)
    np.save(path + "results/results_confidence_recon_all_indi_" + version + ".npy", recon_average_corr_all_indi_new)

    return(recon_average_corr_new_all, recon_average_corr_all_indi_new)


def write_omics_results(path, up_down_list, collected_overlap, recon_average_corr_new_all, headers_all, con_types, data_of_interest): #Todo why only cotinuous data?
     #Todo: since where was no significant hits, couldn't test the function
 
    for i in range(len(con_types)):
        if con_types[i] != data_of_interest:
            for d in collected_overlap:
                n = np.intersect1d(collected_overlap[d], headers_all[i])
                
                with open(path + f"results/{con_types[i]}_" + d.replace(" ", "_") + ".txt", "w") as o:
                    o.write("\n".join(n))  
                
                if con_types[i] in up_down_list:
                     
                    vals = recon_average_corr_new_all[headers_all[i].index(d),np.where(np.isin(con_names,n))[0]]
                    up = n[vals > 0]
                    down = n[vals < 0]
                    with open(path + f"results/{con_types[i]}_up_" + d.replace(" ", "_") + ".txt", "w") as o:
                        o.write("\n".join(up))

                    with open(path + f"results/{con_types[i]}_down_" + d.replace(" ", "_")  + ".txt", "w") as o:
                        o.write("\n".join(down))
                        
                        
def make_files(collected_overlap, groups, con_all, path, recon_average_corr_all_indi_new, 
               con_names, con_dataset_names, drug_h, drug, all_hits, types, version = "v1"):
    
    all_db_names = [item for sublist in con_names for item in sublist]
    ci_dict = {}
    for i,n in enumerate(con_dataset_names):
        ci_collected = []
        sig_drug_names = np.intersect1d(all_hits, n)
        for d in drug_h:
            f = drug_h.index(d)
            recon_data_d = recon_average_corr_all_indi_new[f]
            
            gr = groups[f]
            
            sig_data = recon_data_d[:,np.where(np.isin(all_db_names,sig_drug_names))[0]]
            g = [not (np.all(a_s == types[0]) or (np.all(a_s == [0,0]))) for a_s in drug[gr,f,:]]
            con_tmp = con_all[gr]
            con_tmp = con_tmp[g]
            sig_names_sort = np.array(all_db_names)[np.where(np.isin(all_db_names,sig_drug_names))]
            sig_part_df = pd.DataFrame(sig_data, columns = sig_names_sort)
            sig_part_df = sig_part_df.T
            sig_part_df[np.isnan(sig_part_df)] = 0
            
            ci_all = []
            for j in sig_part_df.index:
                data_vals = np.array(sig_part_df.loc[j,:])
                con_vals = con_tmp[:,all_db_names.index(j)]
                data_vals = data_vals[con_vals != 0]
                ci = stats.t.interval(0.95, len(data_vals)-1, loc=np.nanmean(data_vals), scale=stats.sem(data_vals))
                ci_all.append("%.4f [%.4f, %.4f]"%(np.mean(data_vals), ci[0], ci[1]))
            
            ci_collected.append(ci_all)
        
        ci_collected_df = pd.DataFrame(ci_collected, index = drug_h, columns=sig_names_sort)
        ci_collected_df.T.to_csv(path + "results/" + con_dataset_names[i] + "_ci_sig_" + version +  ".txt", sep = "\t")
        
        
def get_inter_drug_variation(con_names, drug_h, recon_average_corr_all_indi_new, 
                             groups, collected_overlap, drug, con_all, path, types):
    # Inter drug variation 
    all_db_names = [item for sublist in con_names for item in sublist]
    inter_drug_variance = []
    inter_drug_std = []
    for d in drug_h:
        f = drug_h.index(d)
        recon_data_d = recon_average_corr_all_indi_new[f]
        gr = groups[f]
        sig_drug_names = collected_overlap[d]
        sig_data = recon_data_d[:,np.where(np.isin(all_db_names,sig_drug_names))[0]]
        g = [not (np.all(a_s == types[0]) or (np.all(a_s == [0,0]))) for a_s in drug[gr,f,:]]
        con_tmp = con_all[gr]
        con_tmp = con_tmp[g]
        sig_part_df = pd.DataFrame(sig_data, columns = sig_drug_names)
        sig_part_df = sig_part_df.T
        sig_part_df[np.isnan(sig_part_df)] = np.nan
        inter_drug_variance.append(np.nanvar(sig_part_df))
        inter_drug_std.append(np.nanstd(sig_part_df))

    df_indi_var = pd.DataFrame(inter_drug_variance, index=drug_h)
    return(df_indi_var)


def get_drug_similar_each_omics(con_names, con_dataset_names, all_hits, recon_average_corr_new_all, drug_h, version, path):

#     con_names = [con_h, diet_wearables_h, pro_h, targm_h, untargm_h, tran_h, meta_h]
#     con_dataset_names_v1 = ['Clinical continuous', 'Diet and wearables','Proteomics','Targeted metabolomics','Unargeted metabolomics', 'Transcriptomics', 'Metagenomics'] #TODOs define outside maybe
#     con_dataset_names = ['Clinical_continuous', 'Diet_wearables','Proteomics','Targeted_metabolomics','Unargeted_metabolomics', 'Transcriptomics', 'Metagenomics']
    
    con_dataset_names_v1 = con_dataset_names # TODO define the names for plot drawing 
    i = 0
    for n in con_names:
        tmp = np.intersect1d(all_hits, n)
        if len(tmp) == 0:
             continue

        sig_data = recon_average_corr_new_all[:,np.where(np.isin(all_hits,n))[0]]
        sim = cosine_similarity(sig_data)
        corr = pd.DataFrame(sim, columns = drug_h, index = drug_h)
        sig_data = pd.DataFrame(corr, columns = drug_h, index = drug_h)
        g = sns.clustermap(sig_data, cmap=cmap, center=0, xticklabels = True,
                           yticklabels = True, metric='correlation',
                           linewidths=0, row_cluster=True, col_cluster=True, figsize=(10,10))

        g.fig.suptitle(con_dataset_names_v1[i])
        g.fig.subplots_adjust(top=0.9)
        plt.savefig(path + "results/" + con_dataset_names[i] + "_heatmap_" + version + "_all.pdf", format = 'pdf', dpi = 800)
        i += 1

    plt.close('all')


    
def get_best_epoch(results_df):
    
    #Rounding best_epoch to closest 10 (just in case if best_epoch is less than 5 - to closest number)
    round_epoch = lambda x : round(x, 1) if (x >= 5) else round(x, 0)   
    best_epoch = results_df['best_epochs'].mean()
    best_epoch = int(round_epoch(best_epoch))
    
    return(best_epoch)


def get_significant_param_values(results_df, list_of_params, stat_test='t_test'):
    params_dict_to_remove = defaultdict(list)
    
    # Iterating through each of the hypeparameter
    for col in list_of_params:
        
        # Getting each value of the hyperparameter
        unique_values = results_df[col].unique()
        
        # finding which pairs of parameter values do not have significant differences
        insignif_pairs_list = []
        for i in range(len(unique_values)):
            for j in range(i+1, len(unique_values)):
                # Getting the p-value of statistical results by comparing hyperparameter values sets where only the tested hyperparameter is different  
                if stat_test=='t_test':
                    diff_signif = stats.ttest_rel(results_df[results_df[col]==unique_values[i]]['likelihood_test'], results_df[results_df[col]==unique_values[j]]['likelihood_test'])[1]
                elif stat_test=='wilcoxon': 
                    diff_signif = stats.wilcoxon(results_df[results_df[col]==unique_values[i]]['likelihood_test'], results_df[results_df[col]==unique_values[j]]['likelihood_test'])[1]
                    
                if diff_signif < 0.05:
#                     a = tuple([unique_values[i], unique_values[j]])
                    insignif_pairs_list.append(tuple([unique_values[j], unique_values[i]]))
        
        # Removing the parameter that did not show significant differences (in the order from the most insignificant pairs)
        while insignif_pairs_list:
            insignif_values_dict = Counter(value for values_pair in insignif_pairs_list for value in values_pair)
            value_max_occurs = max(insignif_values_dict, key=insignif_values_dict.get)
            insignif_pairs_list = [x for x in insignif_pairs_list if value_max_occurs not in x]
            
            unique_values = np.delete(unique_values, np.where(unique_values == value_max_occurs))
            
            params_dict_to_remove[col].append(value_max_occurs)
    
    # Removing the selected parameters to remove from the dataframe 
    results_df_rm_nonsignifs = pd.DataFrame(results_df)
    for key, value_list in params_dict_to_remove.items():
        for value in value_list:
            results_df_rm_nonsignifs = results_df_rm_nonsignifs[results_df_rm_nonsignifs[key]!=value]

    return(results_df_rm_nonsignifs, params_dict_to_remove)

# def get_significant_param_values(results_df: pd.DataFrame, 
#                                  list_of_params: list, 
#                                  metric_name: str = 'likelihood_test', 
#                                  stat_test: str = 'ttest_rel'):
    
#     params_dict_to_remove = defaultdict(list)

#     # Iterating through each of the hypeparameter
#     for col in list_of_params:

#         # Getting each value of the hyperparameter
#         unique_values = results_df[col].unique()

#         stat_test = getattr(stats, stat_test)  # might be good to catch errors
        
#         # finding which pairs of parameter values do not have significant differences
#         insignif_pairs_list = []
#         for i in range(len(unique_values)):
#             for j in range(i+1, len(unique_values)):
                
#                 # Getting the p-value of statistical results by comparing hyperparameter values sets where only the tested hyperparameter is different

#                 _, pvalue = stat_test(
#                     results_df.loc[results_df[col] == unique_values[i], metric_name],
#                     results_df.loc[results_df[col] == unique_values[j], metric_name])

#                 if pvalue < 0.05:
# #                     a = tuple([unique_values[i], unique_values[j]])
#                     insignif_pairs_list.append(
#                         tuple([unique_values[j], unique_values[i]]))

#         # Removing the parameter that did not show significant differences (in the order from the most insignificant pairs)
#         # strategy: remove the unique value with most insignificant results first, repeat until none is left
#         while insignif_pairs_list:
#             insignif_values_dict = Counter(
#                 value for values_pair in insignif_pairs_list for value in values_pair)
#             value_max_occurs = max(insignif_values_dict,
#                                    key=insignif_values_dict.get)
#             insignif_pairs_list = [
#                 x for x in insignif_pairs_list if value_max_occurs not in x]

#             # unique_values = np.delete(unique_values, np.where(unique_values == value_max_occurs))

#             params_dict_to_remove[col].append(value_max_occurs)

#     # Removing the selected parameters to remove from the dataframe
#     results_df_rm_nonsignifs = pd.DataFrame(results_df)
#     for key, value_list in params_dict_to_remove.items():
#         for value in value_list:
#             results_df_rm_nonsignifs = results_df_rm_nonsignifs[results_df_rm_nonsignifs[key] != value]
#             #isin to compare to a set of values

#     return(results_df_rm_nonsignifs, params_dict_to_remove)

def get_significant_param_values(results_df: pd.DataFrame, 
                                 list_of_params: list, 
                                 metric_name: str = 'likelihood_test', 
                                 stat_test_name: str = 'ttest_rel'):
    
    params_dict_to_remove = defaultdict(list)
    stat_test = getattr(stats, stat_test_name)  # might be good to catch errors
        
    # Iterating through each of the hypeparameter
    for col in list_of_params:

        # Getting each value of the hyperparameter
        unique_values = results_df[col].unique()


        # finding which pairs of parameter values do not have significant differences
        insignif_pairs_list = []
        for i in range(len(unique_values)):
            for j in range(i+1, len(unique_values)):
                
                # Getting the p-value of statistical results by comparing hyperparameter values sets where only the tested hyperparameter is different
                _, pvalue = stat_test(
                    results_df.loc[results_df[col] == unique_values[i], metric_name],
                    results_df.loc[results_df[col] == unique_values[j], metric_name])

                if pvalue < 0.05:
                    insignif_pairs_list.append(tuple([unique_values[j], 
                                                      unique_values[i]]))

        # Removing the parameter that did not show significant differences (in the order from the most insignificant pairs)
        # strategy: remove the unique value with most insignificant results first, repeat until none is left
        while insignif_pairs_list:
            insignif_values_dict = Counter(value for values_pair in insignif_pairs_list for value in values_pair)
            value_max_occurs = max(insignif_values_dict, key=insignif_values_dict.get)
            insignif_pairs_list = [x for x in insignif_pairs_list if value_max_occurs not in x]
            params_dict_to_remove[col].append(value_max_occurs)

    # Removing the selected parameters to remove from the dataframe
    results_df_rm_nonsignifs = pd.DataFrame(results_df)
    for key, value_list in params_dict_to_remove.items():
        results_df_rm_nonsignifs = results_df_rm_nonsignifs[results_df_rm_nonsignifs[key].isin(value_list)]

    return(results_df_rm_nonsignifs, params_dict_to_remove)

def get_sort_list(results_df):
    
    #Gets mean values of test accuracy reconstruction
    results_df['recon_acc_test_mean'] = results_df['recon_acc_test'].map(lambda x: x.mean())
    
    # Sort values by reconstruction accuracy
    results_df = results_df.sort_values('recon_acc_test_mean', ascending=False)
    return (results_df)


def get_length(hyperpars_vals_dict, hyperpar_name):
    lengths = 1
    for key in hyperpars_vals_dict:  
        length = len(hyperpars_vals_dict[key])
        # Adding +1 since we want to check if adding a parameter it overreaches
        if key == hyperpar_name:
            length+=1
        lengths *= length

    return(lengths)

def get_best_params(results_df_sorted, n_combos_opt, hyperpars_names):

    hyperpars_vals_dict = defaultdict(list)
    for index, row in results_df_sorted.iterrows():
        
        for hyperpar_name in hyperpars_names:
            if row[hyperpar_name] not in hyperpars_vals_dict[hyperpar_name]:
                
                length = get_length(hyperpars_vals_dict, hyperpar_name)
                if length <= n_combos_opt:
                    hyperpars_vals_dict[hyperpar_name].append(row[hyperpar_name])
                else:
                    break
    hyperpars_vals_dict = dict(hyperpars_vals_dict)
    return(hyperpars_vals_dict)

def make_and_save_best_reconstruct_params(results_df, hyperparams_names, max_param_combos_to_save):
    
    # Getting the best number of epochs used in further trainings
    best_epoch = get_best_epoch(results_df)
    
    print('Starting calculating the best hyperparameter values for further optimization') 
    
    # Removing insignificant hyperparameter values
    results_df_rm_insignifs, params_dict_to_remove = get_significant_param_values(results_df, hyperparams_names)
    
    #Printing the removed parameter values
    print(f'\nRemoved insignificant parameters:\n {OmegaConf.to_yaml(dict(params_dict_to_remove))}')
    
    # Getting up to n combinations of parameters that will be used further to optimize stability
    results_df_sorted = get_sort_list(results_df_rm_insignifs)
    best_hyperpars_vals_dict = get_best_params(results_df_sorted, max_param_combos_to_save, hyperparams_names)
    best_hyperpars_vals_dict['tuned_num_epochs'] = best_epoch
    
    # Saving the best hyperparameter values (it will overwrite the file if it already exists)
    with open('tuning_stability.yaml', "w") as f:
        OmegaConf.save(OmegaConf.create(best_hyperpars_vals_dict), f)

    #Printing the saved hyper parameter values
    print(f'Saving the best hyperparameter values in tuning_stability.yaml for further optimization: \n{OmegaConf.to_yaml(best_hyperpars_vals_dict)}\n')
    

def get_best_stability_paramset(stability_df, hyperparams_names):
    
    params_to_save = dict()
    
    stability_df_sorted = stability_df.sort_values('difference', ascending=False).iloc[:1]
    for hyperparam in hyperparams_names: 
        params_to_save[hyperparam] = stability_df_sorted[hyperparam].item()
        
    return(params_to_save, stability_df_sorted)

def get_best_4_latent_spaces(results_df_sorted):
    best_latent = []
    for index, row in results_df_sorted.iterrows():
        if row['num_latent'] not in best_latent:
            best_latent.append(int(row['num_latent']))
        if len(best_latent) >= 2:
            break
            
    #Adding two values from both sides 
    best_hypers_diff = max(best_latent) - min(best_latent)
    diff_from_zero = int(min(best_latent)/2)
    
    if min(best_latent) - best_hypers_diff > diff_from_zero:
        best_latent.append(min(best_latent) - best_hypers_diff)
        best_latent.append(max(best_latent) + best_hypers_diff)
    else:
        best_latent.append(min(best_latent) - diff_from_zero)
        best_latent.append(max(best_latent) + diff_from_zero)
    return(best_latent)

def make_and_save_best_stability_params(results_df, hyperparams_names, nepochs):
    
    print('Starting calculating the best hyperparameter values used in further model trainings') 
    
    # Getting best set of hyperparameters
    params_to_save, results_df_sorted = get_best_stability_paramset(results_df, hyperparams_names)
    params_to_save['tuned_num_epochs'] = nepochs

    # Saving best set of hyperparameters    
    with open('training_latent.yaml', "w") as f:
        OmegaConf.save(OmegaConf.create(dict(params_to_save)), f)
        
    # Printing the configuration saved 
    print(f'Saving best hyperparameter values in training_latent.yaml: \n {OmegaConf.to_yaml(dict(params_to_save))}')
    
    # Getting the latent spaces for training_association script and using them with the best hyperparam set
    best_latent = get_best_4_latent_spaces(results_df_sorted)
    params_to_save['num_latent'] = list(best_latent)
    
    # Saving best set of hyperparameters for training_association script
    with open('training_association.yaml', "w") as f:
        OmegaConf.save(OmegaConf.create(dict(params_to_save)), f)

    # Printing the configuration saved 
    print(f'Saving best hyperparameter values in training_association.yaml: \n{OmegaConf.to_yaml(dict(params_to_save))}')



__all__ = ["set_global_seed"]

import random
from typing import Optional

import numpy as np
import torch


def set_global_seed(seed: Optional[int]) -> None:
    if seed is None:
        return
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)



import hydra 
from omegaconf import OmegaConf

from move.conf.schema import MOVEConfig

@hydra.main(config_path="conf", config_name="main")
def main(config: MOVEConfig) -> str:
    print(config.model.continuous_weights)
    print(OmegaConf.to_yaml(config))
    
if __name__ == "__main__":
#     print('hi')
#     pass
    main()




__all__ = ["MOVEDataset", "make_dataloader"]

import numpy as np
import torch
from torch.utils.data import DataLoader, TensorDataset
from typing import Tuple

class MOVEDataset(TensorDataset):
    "Characterizes a dataset for PyTorch"

    def __init__(
        self,
        cat_all: torch.Tensor = None,
        con_all: torch.Tensor = None,
        cat_shapes: list = None,
        con_shapes: list = None,
    ) -> None:
        # Check
        num_samples = None if cat_all is None else cat_all.shape[0]
        if con_all is not None:
            if num_samples is None:
                num_samples = con_all.shape[0]
            elif num_samples != con_all.shape[0]:
                raise ValueError(
                    "Number of samples between categorical and continuous "
                    "datasets must match."
                )
        elif num_samples is None:
            raise ValueError("Categorical and continuous data cannot be both empty.")
        self.num_samples = num_samples
        self.cat_all = cat_all
        self.cat_shapes = cat_shapes
        self.con_all = con_all
        self.con_shapes = con_shapes

    def __len__(self) -> int:
        return self.num_samples

    def __getitem__(self, idx: int) -> Tuple[torch.Tensor]:
        cat_slice = None if self.cat_all is None else self.cat_all[idx]
        con_slice = None if self.con_all is None else self.con_all[idx]
        return cat_slice, con_slice


def concat_cat_list(cat_list):
    cat_shapes = list()
    first = 0

    for cat_d in cat_list:
        cat_shapes.append(cat_d.shape)
        cat_input = cat_d.reshape(cat_d.shape[0], -1)

        if first == 0:
            cat_all = cat_input
            del cat_input
            first = 1
        else:
            cat_all = np.concatenate((cat_all, cat_input), axis=1)

    # Make mask for patients with no measurments
    catsum = cat_all.sum(axis=1)
    mask = catsum > 5
    del catsum
    return cat_shapes, mask, cat_all


def concat_con_list(con_list, mask):
    n_con_shapes = []

    first = 0
    for con_d in con_list:

        n_con_shapes.append(con_d.shape[1])

        if first == 0:
            con_all = con_d
            first = 1
        else:
            con_all = np.concatenate((con_all, con_d), axis=1)

    consum = con_all.sum(axis=1)
    mask &= consum != 0
    del consum
    return n_con_shapes, mask, con_all


def make_dataloader(cat_list=None, con_list=None, batchsize=10, cuda=False):
    """Create a DataLoader for input of each data type - categorical,
    continuous and potentially each omcis set (currently proteomics, target
    metabolomicas, untarget metabolomics and transcriptomics).

    Inputs:
        cat_list: list of categorical input matrix (N_patients x N_variables x N_max-classes)
        con_list: list of normalized continuous input matrix (N_patients x N_variables)
        batchsize: Starting size of minibatches for dataloader
        cuda: Pagelock memory of dataloader (use when using GPU acceleration)

    Outputs:
        DataLoader: An object feeding data to the VAE
    """

    if cat_list is None and con_list is None:
        raise ValueError("At least one type of data must be in the input")

    # Handle categorical data sets
    if not (cat_list is None):
        cat_shapes, mask, cat_all = concat_cat_list(cat_list)

    # else:
    mask = [True] * len(con_list[0])

    # Concetenate con datasetsand make final mask
    if not (con_list is None):
        n_con_shapes, mask, con_all = concat_con_list(con_list, mask)

    # Create dataset
    if not (cat_list is None or con_list is None):
        cat_all = cat_all[mask]
        con_all = con_all[mask]

        cat_all = torch.from_numpy(cat_all)
        con_all = torch.from_numpy(con_all)

        dataset = MOVEDataset(
            con_all=con_all,
            con_shapes=n_con_shapes,
            cat_all=cat_all,
            cat_shapes=cat_shapes,
        )
    elif not (con_list is None):
        con_all = con_all[mask]
        con_all = torch.from_numpy(con_all)
        dataset = MOVEDataset(con_all=con_all, con_shapes=n_con_shapes)
    elif not (cat_list is None):
        cat_all = cat_all[mask]
        cat_all = torch.from_numpy(cat_all)
        dataset = MOVEDataset(cat_all=cat_all, cat_shapes=cat_shapes)
    # Create dataloader
    dataloader = DataLoader(
        dataset=dataset, batch_size=batchsize, drop_last=True, shuffle=True
    )  # Changed num_workers and pin_memory
    return mask, dataloader



__all__ = ["train_model"]

import copy

import hydra
import torch
from torch.utils.data import DataLoader

from move.conf.schema import MOVEConfig
from move.data.dataloaders import make_dataloader
from move.data.io import read_data
from move.models.vae import VAE
from move.training.seed import set_global_seed


@hydra.main(config_path="../conf", config_name="main")
def train_model(config: MOVEConfig):

    if config.seed is not None:
        set_global_seed(config.seed)

    device = torch.device("cuda" if config.training.cuda == True else "cpu")

    cat_list, _, con_list, _ = read_data(config)

    # Making the dataloader
    _, train_loader = make_dataloader(
        cat_list=cat_list, con_list=con_list, batchsize=10
    )  # Added drop_last

    # Make model
    model: VAE = hydra.utils.instantiate(
        config.model,
        continuous_shapes=train_loader.dataset.con_shapes,
        categorical_shapes=train_loader.dataset.cat_shapes,
        _convert_="partial",
    ).to(device)

    kld_w = 0.0
    r = 20 / len(config.training.kld_steps)
    update = 1 + r

    # Lists for saving the results
    losses = list()
    ce = list()
    sse = list()
    KLD = list()

    # Training the model
    for epoch in range(1, config.training.num_epochs + 1):

        if epoch in config.training.kld_steps:
            kld_w = 1 / 20 * update
            update += r

        if epoch in config.training.batch_steps:
            train_loader = DataLoader(
                dataset=train_loader.dataset,
                batch_size=int(
                    train_loader.batch_size * 1.25
                ),  # TODOs whhy train_loader bigger
                shuffle=True,
                drop_last=False,  # Added
                num_workers=train_loader.num_workers,
                pin_memory=train_loader.pin_memory,
            )

        l, c, s, k = model.encoding(train_loader, epoch, config.training.lr, kld_w)

        losses.append(l)
        ce.append(c)
        sse.append(s)
        KLD.append(k)

        best_model = copy.deepcopy(model)

    return best_model, losses, ce, sse, KLD


if __name__ == "__main__":
    train_model()



__all__ = ["VAE"]

import torch
from torch import nn, optim
from torch.utils.data import DataLoader
from typing import List
import numpy as np

class VAE(nn.Module):
    """Variational autoencoder.

    Instantiate with:
        continuous_shapes: shape of the different continuous datasets if any
        categorical_shapes: shape of the different categorical datasets if any
        num_hidden: List of n_neurons in the hidden layers [[200, 200]]
        num_latent: Number of neurons in the latent layer [15]
        beta: Multiply KLD by the inverse of this value [0.0001]
        continuous_weights: list of weights for each continuous dataset
        categorical_weights: list of weights for each categorical dataset
        dropout: Probability of dropout on forward pass [0.2]
        cuda: Use CUDA (GPU accelerated training) [False]

    vae.trainmodel(dataloader, nepochs batchsteps, lrate, logfile, modelfile)
        Trains the model, returning None

    vae.encode(self, data_loader):
        Encodes the data in the data loader and returns the encoded matrix.
    """

    def __init__(
        self,
        categorical_shapes: List[tuple] = None,
        continuous_shapes: List[tuple] = None,
        categorical_weights: List[int] = None,
        continuous_weights: List[int] = None,
        num_hidden: List[int] = [200, 200],
        num_latent: int = 20,
        beta: float = 0.01,
        dropout: float = 0.2,
        cuda: bool = False,
    ):
        if num_latent < 1:
            raise ValueError(f"Minimum 1 latent unit. Input was {num_latent}.")

        if beta <= 0:
            raise ValueError("Beta must be greater than zero.")

        if not (0 <= dropout < 1):
            raise ValueError("Dropout must be between zero and one.")

        if continuous_shapes is None and categorical_shapes is None:
            raise ValueError("Shapes of the input data must be provided.")

        num_categorical = sum([int.__mul__(*shape[1:]) for shape in categorical_shapes])
        num_continuous = sum(continuous_shapes)

        self.input_size = 0
        if not (num_continuous is None or continuous_shapes is None):
            self.num_continuous = num_continuous
            self.input_size += self.num_continuous
            self.continuous_shapes = continuous_shapes

            if not (continuous_weights is None):
                self.continuous_weights = continuous_weights
                if not len(continuous_shapes) == len(continuous_weights):
                    raise ValueError(
                        "Number of continuous weights must be the same as"
                        " number of continuous datasets"
                    )
        else:
            self.num_continuous = None

        if not (num_categorical is None or categorical_shapes is None):
            self.num_categorical = num_categorical
            self.input_size += self.num_categorical
            self.categorical_shapes = categorical_shapes

            if not (categorical_weights is None):
                self.categorical_weights = categorical_weights
                if not len(categorical_shapes) == len(categorical_weights):
                    raise ValueError(
                        "Number of categorical weights must be the same as"
                        " number of categorical datasets"
                    )
        else:
            self.num_categorical = None

        super(VAE, self).__init__()

        # Initialize simple attributes
        self.beta = beta
        self.num_hidden = num_hidden
        self.num_latent = num_latent
        self.dropout = dropout

        self.device = torch.device("cuda" if cuda == True else "cpu")

        # Activation functions
        self.relu = nn.LeakyReLU()
        self.softplus = nn.Softplus()
        self.sigmoid = nn.Sigmoid()
        self.log_softmax = nn.LogSoftmax(dim=1)
        self.dropoutlayer = nn.Dropout(p=self.dropout)

        # Initialize lists for holding hidden layers
        self.encoderlayers = nn.ModuleList()
        self.encodernorms = nn.ModuleList()
        self.decoderlayers = nn.ModuleList()
        self.decodernorms = nn.ModuleList()

        ### Layers
        # Hidden layers
        for nin, nout in zip([self.input_size] + self.num_hidden, self.num_hidden):
            self.encoderlayers.append(nn.Linear(nin, nout))
            self.encodernorms.append(nn.BatchNorm1d(nout))

        # Latent layers
        self.mu = nn.Linear(self.num_hidden[-1], self.num_latent)  # mu layer
        self.var = nn.Linear(self.num_hidden[-1], self.num_latent)  # logvariance layer

        # Decoding layers
        for nin, nout in zip(
            [self.num_latent] + self.num_hidden[::-1], self.num_hidden[::-1]
        ):
            self.decoderlayers.append(nn.Linear(nin, nout))
            self.decodernorms.append(nn.BatchNorm1d(nout))

        # Reconstruction - output layers
        self.out = nn.Linear(self.num_hidden[0], self.input_size)  # to output

    def encode(self, x):
        # Hidden layers
        for encoderlayer, encodernorm in zip(self.encoderlayers, self.encodernorms):
            x = encoderlayer(x)
            x = self.relu(x)
            x = self.dropoutlayer(x)
            x = encodernorm(x)

        return self.mu(x), self.var(x)

    def reparameterize(self, mu, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)

        return eps.mul(std).add_(mu)

    def decompose_categorical(self, reconstruction):
        cat_tmp = reconstruction.narrow(1, 0, self.num_categorical)

        # handle soft max for each categorical dataset
        cat_out = []
        pos = 0
        for cat_shape in self.categorical_shapes:
            cat_dataset = cat_tmp[:, pos : (cat_shape[1] * cat_shape[2] + pos)]

            cat_out_tmp = cat_dataset.view(
                cat_dataset.shape[0], cat_shape[1], cat_shape[2]
            )
            cat_out_tmp = cat_out_tmp.transpose(1, 2)
            cat_out_tmp = self.log_softmax(cat_out_tmp)

            cat_out.append(cat_out_tmp)
            pos += cat_shape[1] * cat_shape[2]

        return cat_out

    def decode(self, x):
        for decoderlayer, decodernorm in zip(self.decoderlayers, self.decodernorms):
            x = decoderlayer(x)
            x = self.relu(x)
            x = self.dropoutlayer(x)
            x = decodernorm(x)

        reconstruction = self.out(x)

        # Decompose reconstruction to categorical and continuous variables
        # if both types are in the input
        if not (self.num_categorical is None or self.num_continuous is None):
            cat_out = self.decompose_categorical(reconstruction)
            con_out = reconstruction.narrow(
                1, self.num_categorical, self.num_continuous
            )
        elif not (self.num_categorical is None):
            cat_out = self.decompose_categorical(reconstruction)
            con_out = None
        elif not (self.num_continuous is None):
            cat_out = None
            con_out = reconstruction.narrow(1, 0, self.num_continuous)

        return cat_out, con_out

    def forward(self, tensor):
        mu, logvar = self.encode(tensor)
        z = self.reparameterize(mu, logvar)
        cat_out, con_out = self.decode(z)

        return cat_out, con_out, mu, logvar

    def calculate_cat_error(self, cat_in, cat_out):
        batch_size = cat_in.shape[0]

        # calcualte target values for all cat datasets
        count = 0
        cat_errors = []
        pos = 0
        for cat_shape in self.categorical_shapes:
            cat_dataset = cat_in[:, pos : (cat_shape[1] * cat_shape[2] + pos)]

            cat_dataset = cat_dataset.view(cat_in.shape[0], cat_shape[1], cat_shape[2])
            cat_target = cat_dataset
            cat_target = cat_target.argmax(2)
            cat_target[cat_dataset.sum(dim=2) == 0] = -1
            cat_target = cat_target.to(self.device)

            # Cross entropy loss for categroical
            loss = nn.NLLLoss(reduction="sum", ignore_index=-1)
            cat_errors.append(
                loss(cat_out[count], cat_target) / (batch_size * cat_shape[1])
            )
            count += 1
            pos += cat_shape[1] * cat_shape[2]

        cat_errors = torch.stack(cat_errors)
        return cat_errors

    def calculate_con_error(self, con_in, con_out, loss):
        batch_size = con_in.shape[0]
        total_shape = 0
        con_errors = []
        for s in self.continuous_shapes:
            c_in = con_in[:, total_shape : (s + total_shape - 1)]
            c_re = con_out[:, total_shape : (s + total_shape - 1)]
            error = loss(c_re, c_in) / batch_size
            con_errors.append(error)
            total_shape += s

        con_errors = torch.stack(con_errors)
        con_errors = con_errors / torch.Tensor(self.continuous_shapes)
        MSE = torch.sum(con_errors * torch.Tensor(self.continuous_weights))
        return MSE

    # Reconstruction + KL divergence losses summed over all elements and batch
    def loss_function(self, cat_in, cat_out, con_in, con_out, mu, logvar, kld_w):
        MSE = 0
        CE = 0
        # calculate loss for catecorical data if in the input
        if not (cat_out is None):
            cat_errors = self.calculate_cat_error(cat_in, cat_out)
            if not (self.categorical_weights is None):
                CE = torch.sum(cat_errors * torch.Tensor(self.categorical_weights))
            else:
                CE = torch.sum(cat_errors) / len(cat_errors)

        # calculate loss for continuous data if in the input
        if not (con_out is None):
            batch_size = con_in.shape[0]
            # Mean square error loss for continauous
            loss = nn.MSELoss(reduction="sum")
            # set missing data to 0 to remove any loss these would provide
            con_out[con_in == 0] == 0

            # include different weights for each omics dataset
            if not (self.continuous_weights is None):
                MSE = self.calculate_con_error(con_in, con_out, loss)
            else:
                MSE = loss(con_out, con_in) / (batch_size * self.num_continuous)

        # see Appendix B from VAE paper:
        # Kingma and Welling. Auto-Encoding Variational Bayes. ICLR, 2014
        # https://arxiv.org/abs/1312.6114
        # 0.5 * sum(1 + log(sigma^2) - mu^2 - sigma^2)
        KLD = -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp()) / (batch_size)

        KLD_weight = self.beta * kld_w
        loss = CE + MSE + KLD * KLD_weight

        return loss, CE, MSE, KLD * KLD_weight

    def encoding(self, train_loader, epoch, lrate, kld_w):
        self.train()
        train_loss = 0
        log_interval = 50

        optimizer = optim.Adam(self.parameters(), lr=lrate)

        epoch_loss = 0
        epoch_kldloss = 0
        epoch_sseloss = 0
        epoch_bceloss = 0

        for batch_idx, (cat, con) in enumerate(train_loader):
            # Move input to GPU if requested
            cat = cat.to(self.device)
            con = con.to(self.device)

            if not (self.num_categorical is None or self.num_continuous is None):
                tensor = torch.cat((cat, con), 1)
            elif not (self.num_categorical is None):
                tensor = cat
            elif not (self.num_continuous is None):
                tensor = con

            optimizer.zero_grad()

            cat_out, con_out, mu, logvar = self(tensor)

            loss, bce, sse, kld = self.loss_function(
                cat, cat_out, con, con_out, mu, logvar, kld_w
            )
            loss.backward()

            epoch_loss += loss.data.item()
            epoch_kldloss += kld.data.item()

            if not (self.num_continuous is None):
                epoch_sseloss += sse.data.item()

            if not (self.num_categorical is None):
                epoch_bceloss += bce.data.item()

            optimizer.step()

        print(
            "\tEpoch: {}\tLoss: {:.6f}\tCE: {:.7f}\tSSE: {:.6f}\t"
            "KLD: {:.4f}\tBatchsize: {}".format(
                epoch,
                epoch_loss / len(train_loader),
                epoch_bceloss / len(train_loader),
                epoch_sseloss / len(train_loader),
                epoch_kldloss / len(train_loader),
                train_loader.batch_size,
            )
        )
        return (
            epoch_loss / len(train_loader),
            epoch_bceloss / len(train_loader),
            epoch_sseloss / len(train_loader),
            epoch_kldloss / len(train_loader),
        )

    def make_cat_recon_out(self, length):
        cat_total_shape = 0
        for cat_shape in self.categorical_shapes:
            cat_total_shape += cat_shape[1]

        cat_class = torch.empty((length, cat_total_shape)).int()
        cat_recon = torch.empty((length, cat_total_shape)).int()
        return cat_class, cat_recon, cat_total_shape

    def get_cat_recon(self, batch, cat_total_shape, cat, cat_out):
        count = 0
        cat_out_class = torch.empty((batch, cat_total_shape)).int()
        cat_target = torch.empty((batch, cat_total_shape)).int()
        pos = 0
        shape_1 = 0
        for cat_shape in self.categorical_shapes:
            # Get input categorical data
            cat_in_tmp = cat[:, pos : (cat_shape[1] * cat_shape[2] + pos)]
            cat_in_tmp = cat_in_tmp.view(cat.shape[0], cat_shape[1], cat_shape[2])

            # Calculate target values for input
            cat_target_tmp = cat_in_tmp
            cat_target_tmp = torch.argmax(cat_target_tmp.detach(), dim=2)
            cat_target_tmp[cat_in_tmp.sum(dim=2) == 0] = -1
            cat_target[:, shape_1 : (cat_shape[1] + shape_1)] = cat_target_tmp#.numpy()

            # Get reconstructed categorical data
            cat_out_tmp = cat_out[count]
            cat_out_tmp = cat_out_tmp.transpose(1, 2)
            cat_out_class[:, shape_1 : (cat_shape[1] + shape_1)] = torch.argmax(
                cat_out_tmp, dim=2
            )#.numpy()

            # make counts for next dataset
            pos += cat_shape[1] * cat_shape[2]
            shape_1 += cat_shape[1]
            count += 1

        cat_target = cat_target.numpy()
        cat_out_class = cat_out_class.numpy()
        
        return cat_out_class, cat_target

    @torch.no_grad()
    def latent(self, dataloader: DataLoader, kld_weight: float):
        self.eval()
        test_loss = 0
        test_likelihood = 0

        num_samples = dataloader.dataset.num_samples

        latent = torch.empty((num_samples, self.num_latent))
        latent_var = torch.empty((num_samples, self.num_latent))

        # reconstructed output
        if not (self.num_categorical is None):
            cat_class, cat_recon, cat_total_shape = self.make_cat_recon_out(num_samples)
        else:
            cat_class = None
            cat_recon = None

        con_recon = (
            None
            if self.num_continuous is None
            else torch.empty((num_samples, self.num_continuous))
        )

        row = 0
        for (cat, con) in dataloader:
            cat = cat.to(self.device)
            con = con.to(self.device)

            # get dataset
            if not (self.num_categorical is None or self.num_continuous is None):
                tensor = torch.cat((cat, con), 1)
            elif not (self.num_categorical is None):
                tensor = cat
            elif not (self.num_continuous is None):
                tensor = con

            # Evaluate
            cat_out, con_out, mu, logvar = self(tensor)

            mu = mu.to(self.device)
            logvar = logvar.to(self.device)
            batch = len(mu)

            loss, bce, sse, _ = self.loss_function(
                cat, cat_out, con, con_out, mu, logvar, kld_weight
            )
            test_likelihood += bce + sse
            test_loss += loss.data.item()

            if not (self.num_categorical is None):
                cat_out_class, cat_target = self.get_cat_recon(
                    batch, cat_total_shape, cat, cat_out
                )
                cat_recon[row : row + len(cat_out_class)] = torch.Tensor(cat_out_class)
                cat_class[row : row + len(cat_target)] = torch.Tensor(cat_target)

            if not (self.num_continuous is None):
                con_recon[row : row + len(con_out)] = con_out

            latent_var[row : row + len(logvar)] = logvar
            latent[row : row + len(mu)] = mu
            row += len(mu)

        test_loss /= len(dataloader)
        print("====> Test set loss: {:.4f}".format(test_loss))
        
        latent = latent.numpy()
        latent_var = latent_var.numpy()
        cat_recon = cat_recon.numpy()
        cat_class = cat_class.numpy()
        con_recon = con_recon.numpy()
                
        
        assert row == num_samples
        return (
            latent,
            latent_var,
            cat_recon,
            cat_class,
            con_recon,
            test_loss,
            test_likelihood,
        )



# Load functions
import hydra 
from move.conf.schema import MOVEConfig

from move.training.train import train_model
from move.utils.data_utils import get_data, merge_configs
from move.utils.visualization_utils import embedding_plot_discrete, embedding_plot_float, visualize_training, plot_reconstruction_distribs, visualize_embedding, plot_categorical_importance, plot_continuous_importance
from move.utils.analysis import get_latents, calc_categorical_reconstruction_acc, calc_continuous_reconstruction_acc, get_embedding, get_pearsonr, get_feature_importance_categorical, get_feature_importance_continuous, save_feat_results, get_feat_importance_on_weights 

import numpy as np

@hydra.main(config_path="../conf", config_name="main")
def main(base_config: MOVEConfig): 
    
    # Merging the user defined data.yaml, model.yaml and training_latent.yaml 
    # with the base_config to override it.
    print('Overriding the default config with configs from data.yaml, model.yaml and training_latent.yaml')
    cfg = merge_configs(base_config=base_config, 
                config_types=['data', 'model', 'training_latent'])
    
    #Getting the variables used in the notebook
    raw_data_path = cfg.data.raw_data_path
    interim_data_path = cfg.data.interim_data_path
    processed_data_path = cfg.data.processed_data_path 
    data_of_interest = cfg.data.data_of_interest
    categorical_names = cfg.data.categorical_names
    continuous_names = cfg.data.continuous_names
    categorical_weights = cfg.data.categorical_weights
    continuous_weights = cfg.data.continuous_weights  
    features_to_visualize = cfg.data.data_features_to_visualize_notebook4
    
    seed = cfg.model.seed
    cuda = cfg.model.cuda   
    lrate = cfg.model.lrate
    kld_steps = cfg.model.kld_steps
    batch_steps = cfg.model.batch_steps

    nHiddens = cfg.training_latent.num_hidden
    nLatents = cfg.training_latent.num_latent
    nLayers = cfg.training_latent.num_layers
    nDropout = cfg.training_latent.dropout
    nBeta = cfg.training_latent.beta
    batch_sizes = cfg.training_latent.batch_sizes
    nepochs = cfg.training_latent.tuned_num_epochs 
    
    epochs = range(1, nepochs + 1)    
    
    #Getting the data
    cat_list, con_list, cat_names, con_names, headers_all, drug, drug_h = get_data(raw_data_path, interim_data_path, categorical_names, continuous_names, data_of_interest)
    
    # Checking if all input features selected for visualization were in headers_all
    for feature in features_to_visualize:
        if feature not in np.concatenate(headers_all):
            raise ValueError(f"{feature} is not in the headers_all list. It could have been it was not among the features of the input dataset or was filtered out during data processing")
    
    # Training the model 
    print('Beginning training the model.\n')
    best_model, losses, ce, sse, KLD, train_loader, mask, kld_w, cat_shapes, con_shapes, best_epoch = train_model(cat_list, con_list, categorical_weights, continuous_weights, batch_sizes, nHiddens, nLayers, nLatents, nBeta, nDropout, cuda, kld_steps, batch_steps, nepochs, lrate, seed, test_loader=None, patience=None, early_stopping=False)
    print('\nFinished training the model.')
    
    # Visualizing the training
    visualize_training(processed_data_path, losses, ce, sse, KLD, epochs)
    
    # Getting the reconstruction results
    latent, latent_var, cat_recon, cat_class, con_recon, loss, likelihood = get_latents(best_model, train_loader, 1)
    cat_total_recon = calc_categorical_reconstruction_acc(cat_shapes, cat_class, cat_recon)
    all_values = calc_continuous_reconstruction_acc(con_shapes, con_recon, train_loader)
    
    # Plotting the reconstruction distributions   
    plot_reconstruction_distribs(processed_data_path, cat_total_recon, all_values)
    
    # Getting the embeddings
    print('\n Getting the embeddings.')    
    embedding = get_embedding(processed_data_path, latent)
    
    # Visualizing the embedding of three example features
    for feature in features_to_visualize:
        visualize_embedding(processed_data_path, feature, embedding, 
                            mask, cat_list, con_list, cat_names, con_names)
        
    # Getting pearson correlations of two example features
    for feature in features_to_visualize:
        spear_corr = get_pearsonr(feature, embedding, cat_list, con_list, cat_names, con_names)
        print(f"Pearson correlation for the 1st embedding dim of {feature}: {round(spear_corr[0][0], 3)}, p-value={round(spear_corr[0][1], 3)}")
        print(f"Pearson correlation for the 2nd embedding dim of {feature}: {round(spear_corr[1][0], 3)}, p-value={round(spear_corr[1][1], 3)}")
        
    # Getting features importance measures
    all_diffs, all_diffs_cat_np, sum_diffs_cat_np, sum_diffs_cat_abs_np,\
        total_diffs_cat_np = get_feature_importance_categorical(best_model, train_loader, latent)
    all_diffs_con_np, sum_diffs_con_np, sum_diffs_con_abs_np,\
            total_diffs_con_np = get_feature_importance_continuous(best_model, train_loader, mask, latent)

    # Saving features importance measure results 
    save_feat_results(processed_data_path, all_diffs_cat_np, sum_diffs_cat_np, sum_diffs_cat_abs_np, total_diffs_cat_np, 
                 all_diffs_con_np, sum_diffs_con_np,sum_diffs_con_abs_np, total_diffs_con_np)
    
    # Plotting categorical importance measures
    plot_categorical_importance(path=processed_data_path,
                            sum_diffs=sum_diffs_cat_np,
                            cat_list=cat_list,
                            feature_names=cat_names,
                            fig_name='importance_SHAP_cat')
    
    # Plotting continuous importance measures
    plot_continuous_importance(path=processed_data_path,
                           train_loader=train_loader,
                           sum_diffs=sum_diffs_con_np,
                           feature_names=con_names,
                           fig_name='importance_SHAP_con')
    
    # Getting feature importance on weights
    get_feat_importance_on_weights(processed_data_path, best_model, train_loader, cat_names, con_names)


if __name__ == "__main__":
    main()


__all__ = ["read_config"]

from pathlib import Path
from typing import Union

import hydra
import numpy as np
from omegaconf import OmegaConf

from move import conf
from move.conf.schema import MOVEConfig


def read_config(filepath: Union[str, Path]) -> MOVEConfig:
    """Composes configuration for the MOVE framework.

    Parameters
    ----------
    filepath : Union[str, Path]
        Path to YAML configuration file

    Returns
    -------
    move.conf.MOVEConfig
    """
    with hydra.initialize_config_module(conf.__name__):
        base_config = hydra.compose("main")

    user_config = OmegaConf.load(filepath)
    return OmegaConf.merge(base_config, user_config)


# NOTE: Adapted from Notebook #4

# Functions for reading data
def read_cat(file):
    data = np.load(file)
    data = data.astype(np.float32)
    data_input = data.reshape(data.shape[0], -1)
    return data, data_input  # ???: Is second output necessary?


def read_con(file):
    data = np.load(file)
    data = data.astype(np.float32)
    data[np.isnan(data)] = 0
    consum = data.sum(axis=0)
    mask_col = consum != 0
    data = data[:, mask_col]
    return data, mask_col


def read_header(file, mask=None, start=1):
    with open(file, "r") as f:
        h = f.readline().rstrip().split("\t")[start:]
    if not mask is None:
        h = np.array(h)
        h = h[mask]
    return h


def read_data(config: MOVEConfig):
    categorical_vars = []
    continuous_vars = []

    categorical_headers = []
    continuous_headers = []

    output_path = Path(config.data.interim_data_path)

    for var_list, header_list, input_configs in [
        (categorical_vars, categorical_headers, config.data.categorical_inputs),
        (continuous_vars, continuous_headers, config.data.continuous_inputs),
    ]:
        for input_config in input_configs:
            var, _ = read_cat(output_path / f"{input_config.name}.npy")
            header = read_header(output_path / f"{input_config.name}.tsv")
            var_list.append(var)
            header_list += header

    return categorical_vars, categorical_headers, continuous_vars, continuous_headers



__license__ = "MIT"
__version__ = (1, 0, 5)
__all__ = ["analysis", "conf", "data", "models", "training", "VAE"]

from move import analysis, conf, models, training, data
from move.models.vae import VAE


# Import functions
import hydra 
from move.conf.schema import MOVEConfig

from move.utils.data_utils import generate_file, merge_configs 

@hydra.main(config_path="../conf", config_name="main")
def main(base_config: MOVEConfig):
    
    # Merging the user defined data.yaml, model.yaml and tuning_reconstruction.yaml 
    # with the base_config to override it.
    print('Overriding the default configuration with configuration from data.yaml')
    cfg = merge_configs(base_config=base_config, 
                        config_types=['data'])
    
    # Getting the variables used in the notebook
    raw_data_path = cfg.data.raw_data_path
    interim_data_path = cfg.data.interim_data_path
    ids_file_name = cfg.data.ids_file_name
    na_encoding = cfg.data.na_value
    categorical_names = cfg.data.categorical_names
    continuous_names = cfg.data.continuous_names    
    
    # Encoding categorical data
    print('Encoding categorical data')
    for cat_data in categorical_names:
        generate_file('categorical', raw_data_path, interim_data_path, cat_data, ids_file_name, na_encoding)
        print(f'  Encoded {cat_data}')
    
    # Encoding continuous data 
    print('Encoding continuous data')
    for con_data in continuous_names:
        generate_file('continuous', raw_data_path, interim_data_path, con_data, ids_file_name, na_encoding)    
        print(f'  Encoded {con_data}')

if __name__ == "__main__":
    main()


# Import functions
import hydra 
from move.conf.schema import MOVEConfig

from move.training.train import optimize_reconstruction
from move.utils.data_utils import get_data, merge_configs
from move.utils.visualization_utils import visualize_likelihood, visualize_recon_acc
from move.utils.analysis import make_and_save_best_reconstruct_params

@hydra.main(config_path="../conf", config_name="main")
def main(base_config: MOVEConfig): 
    
    # Merging the user defined data.yaml, model.yaml and tuning_reconstruction.yaml 
    # with the base_config to override it.
    print('Overriding the default config with configs from data.yaml, model.yaml and tuning_reconstruction.yaml')
    cfg = merge_configs(base_config=base_config, 
                        config_types=['data', 'model', 'tuning_reconstruction'])
    
    # Getting the variables used in the notebook
    
    raw_data_path = cfg.data.raw_data_path
    interim_data_path = cfg.data.interim_data_path
    processed_data_path = cfg.data.processed_data_path
    data_of_interest = cfg.data.data_of_interest
    categorical_names = cfg.data.categorical_names
    continuous_names = cfg.data.continuous_names
    categorical_weights = cfg.data.categorical_weights
    continuous_weights = cfg.data.continuous_weights
    
    seed = cfg.model.seed
    cuda = cfg.model.cuda
    nepochs = cfg.model.num_epochs
    kld_steps = cfg.model.kld_steps
    batch_steps = cfg.model.batch_steps
    patience = cfg.model.patience
    lrate = cfg.model.lrate
    
    nHiddens = cfg.tuning_reconstruction.num_hidden
    nLatents = cfg.tuning_reconstruction.num_latent
    nLayers = cfg.tuning_reconstruction.num_layers
    nDropout = cfg.tuning_reconstruction.dropout
    nBeta = cfg.tuning_reconstruction.beta
    batch_sizes = cfg.tuning_reconstruction.batch_sizes
    repeat = cfg.tuning_reconstruction.repeats  
    max_param_combos_to_save = cfg.tuning_reconstruction.max_param_combos_to_save
    
    #Getting the data
    cat_list, con_list, cat_names, con_names, headers_all, drug, drug_h = get_data(raw_data_path, interim_data_path, categorical_names, continuous_names, data_of_interest)

    #Performing hyperparameter tuning

    likelihood_tests, recon_acc_tests, recon_acc, results_df = optimize_reconstruction(nHiddens, nLatents, 
                                                                                       nLayers, nDropout, 
                                                                                       nBeta, batch_sizes, 
                                                                                       nepochs, repeat, 
                                                                                       lrate, kld_steps, 
                                                                                       batch_steps, patience, 
                                                                                       cuda, processed_data_path, 
                                                                                       cat_list, con_list,
                                                                                       continuous_weights, 
                                                                                       categorical_weights,
                                                                                       seed)
    

    # Visualizing the data
    try:
        visualize_likelihood(processed_data_path, nLayers, nHiddens, nDropout, nBeta, nLatents, likelihood_tests)
        visualize_recon_acc(processed_data_path, nLayers, nHiddens, nDropout, nBeta, nLatents, recon_acc_tests, 'test')
        visualize_recon_acc(processed_data_path, nLayers, nHiddens, nDropout, nBeta, nLatents, recon_acc, 'train')  
        print('Visualizing the hyperparameter tuning results\n')
    except:
        print('Could not visualize the results\n')

    # Getting and saving the best n hyperparameter set value combinations for further optimisation 
    hyperparams_names = ['num_hidden','num_latent', 'num_layers', 'dropout', 'beta', 'batch_sizes']
    make_and_save_best_reconstruct_params(results_df, hyperparams_names, max_param_combos_to_save)

    return()


if __name__ == "__main__":
    main()


__all__ = ["MOVEConfig"]

from dataclasses import dataclass

from hydra.core.config_store import ConfigStore
from omegaconf import OmegaConf
from typing import List

@dataclass
class InputConfig:
    name: str
    weight: float

@dataclass
class DataConfig:
    user_conf: str
    na_value: str
    raw_data_path: str
    interim_data_path: str
    processed_data_path: str
    version: str
    ids_file_name: str
    categorical_inputs: List[InputConfig]
    continuous_inputs: List[InputConfig]
    data_of_interest: str
    categorical_names: List[str]
    continuous_names: List[str]
    categorical_weights: List[float]
    continuous_weights: List[float]
    data_features_to_visualize_notebook4: List[str]
    write_omics_results_notebook5: List[str]
                
@dataclass
class ModelConfig:
    _target_: str
    cuda: bool

@dataclass
class VAEConfig(ModelConfig):
    user_conf: str
    seed: int
    cuda: bool
    lrate: float
    num_epochs: int
    patience: int
    kld_steps: List[int]
    batch_steps: List[int]
        
@dataclass
class TuningReconstructionConfig:
    user_config: str
    num_hidden: List[int]
    num_latent: List[int]
    num_layers: List[int]
    beta: List[float]
    dropout: List[float]
    batch_sizes: List[int]
    repeats: int
    max_param_combos_to_save: int

@dataclass
class TuningStabilityConfig:
    user_config: str
    num_hidden: List[int]
    num_latent: List[int]
    num_layers: List[int]
    beta: List[float]
    dropout: List[float]
    batch_sizes: List[int]
    repeat: int  
    tuned_num_epochs: int
        
@dataclass
class TrainingLatentConfig:
    user_config: str
    num_hidden: int
    num_latent: int
    num_layers: int
    dropout: float
    beta: float
    batch_sizes: int 
    tuned_num_epochs: int
        
@dataclass
class TrainingAssociationConfig:
    user_config: str
    num_hidden: int
    num_latent: List[int]
    num_layers: int
    dropout: float
    beta: float
    batch_sizes: int 
    repeats: int 
    tuned_num_epochs: int

@dataclass
class MOVEConfig:
    data: DataConfig
    model: VAEConfig
    tune_reconstruction: TuningReconstructionConfig
    tune_stability: TuningStabilityConfig
    train_latent: TrainingLatentConfig
    train_association: TrainingAssociationConfig
    name: str


def extract_weights(configs: List[InputConfig]) -> List[int]:
    """Extracts the weights from a list of input configs."""
    return [item.weight for item in configs]

def extract_names(configs: List[InputConfig]) -> List[int]:
    """Extracts the weights from a list of input configs."""
    return [item.name for item in configs]

# Store config schema
cs = ConfigStore.instance()
cs.store(name="config", node=MOVEConfig)

# Register custom resolvers
OmegaConf.register_new_resolver("weights", extract_weights)
OmegaConf.register_new_resolver("names", extract_names)



# Import functions
import os
import torch
import numpy as np
import copy
import pandas as pd 

from torch.utils.data import DataLoader

import umap.umap_ as umap
import random
import itertools 

from move.models import vae
from move.utils import dataloaders
from move.utils.model_utils import get_latent, cal_recon, cal_cat_recon, cal_con_recon, get_baseline, change_drug, cal_sig_hits, correction_new, get_start_end_positions 
from move.utils.data_utils import initiate_default_dicts
from move.utils.seed import set_global_seed


def optimize_reconstruction(nHiddens, nLatents, nLayers, nDropout, nBeta, batch_sizes, nepochs, repeat, lrate, kldsteps, batchsteps, patience, cuda, processed_data_path, cat_list, con_list, continuous_weights, categorical_weights, seed):
    """
    Performs hyperparameter tuning in terms of reconstruction
    
    inputs:
    returns:
    """ 

    # Preparing the data
    isExist = os.path.exists(processed_data_path + 'hyperparameters/')
    if not isExist:
        os.makedirs(processed_data_path + 'hyperparameters/')

    # Divide into test and training set
    npatient = cat_list[0].shape[0]
    train = random.sample(range(npatient), int(npatient*0.90))
    test = list(set(range(npatient)) - set(train))

    cat_list_train = []
    con_list_train = []
    cat_list_test = []
    con_list_test = []

    # Splitting into train and test sets
    for cat_data in cat_list:
        cat_list_train.append(cat_data[train])
        cat_list_test.append(cat_data[test])
    for con_data in con_list:
        con_list_train.append(con_data[train])
        con_list_test.append(con_data[test]) 

    # The selections are are saved
    np.save(processed_data_path + "hyperparameters/train1.npy", np.array(train))
    np.save(processed_data_path + "hyperparameters/test1.npy", np.array(test))

     # Create objects to save results
    latents, con_recons, cat_recons, recon_acc, loss_train,\
    likelihoods, best_epochs, latents_tests, con_recons_tests,\
    cat_recons_tests, recon_acc_tests, loss_tests, likelihood_tests = initiate_default_dicts(0, 13)

    # Getting the data
    mask_test, test_loader = dataloaders.make_dataloader(cat_list=cat_list_test, con_list=con_list_test, batchsize=1)
    test_loader = DataLoader(dataset=test_loader.dataset, batch_size=1, drop_last=False, shuffle=False) #, num_workers=1, pin_memory=test_loader.pin_memory

    results_df = []
    
    print('Beginning the hyperparameter tuning for reconstruction.\n')
    iters = itertools.product(nHiddens, nLatents, nLayers, nDropout, nBeta, batch_sizes, range(repeat))
    for nHidden, nLatent, nl, drop, b, batch_size, r in iters:
        combi = str([nHidden] * nl) + "+" + str(nLatent) + ", drop: " + str(drop) +", b: " + str(b) + ", batch: " + str(batch_size)

        best_model, loss, ce, sse, KLD, train_loader, mask, kld_w, cat_shapes, con_shapes, best_epoch = train_model(cat_list, con_list, categorical_weights, continuous_weights, batch_size, nHidden, nl, nLatent, b, drop, cuda, kldsteps, batchsteps, nepochs, lrate, seed, test_loader, patience, early_stopping=True)   


        # get results
        con_recon, train_test_loader, latent, latent_var, cat_recon, cat_class, loss, likelihood, latent_test, latent_var_test, cat_recon_test, cat_class_test, con_recon_test, loss_test, likelihood_test = get_latent(best_model, train_loader, test_loader, kld_w)

        # Calculate reconstruction
        cat_true_recon,true_recon,cat_true_recon_test,true_recon_test = cal_recon(cat_shapes, cat_recon, cat_class, train_loader, con_recon, con_shapes, cat_recon_test, cat_class_test, test_loader, con_recon_test)

        # Save output
        recon_acc[combi].append(cat_true_recon + true_recon)
        latents[combi].append(latent)
        con_recons[combi].append(con_recon)
        cat_recons[combi].append(cat_recon)
        loss_train[combi].append(loss)
        likelihoods[combi].append(likelihood)
        best_epochs[combi].append(best_epoch)

        recon_acc_tests[combi].append(cat_true_recon_test + true_recon_test)
        latents_tests[combi].append(latent_test)
        con_recons_tests[combi].append(con_recon_test)
        cat_recons_tests[combi].append(cat_recon_test)
        loss_tests[combi].append(loss_test)
        likelihood_tests[combi].append(likelihood_test)
        
        results_df.append({
            'num_hidden': nHidden,
            'num_latent': nLatent,
            'num_layers': nl,
            'dropout': drop,
            'beta': b,
            'batch_sizes': batch_size, 
            'repeats': r,
            'recon_acc': cat_true_recon + true_recon, 
            'latents': np.array(latent), 
            'con_recons': np.array(con_recon),
            'cat_recons': np.array(cat_recon),
            'loss_train': np.array(loss),
            'likelihoods': np.array(likelihood),
            'best_epochs': np.array(best_epoch),
            'recon_acc_test': np.array(cat_true_recon_test + true_recon_test),
            'latents_test': np.array(latent_test),
            'con_recons_test': np.array(con_recon_test),
            'cat_recons_test': np.array(cat_recon_test),
            'loss_test': np.array(loss_test),
            'likelihood_test': np.array(likelihood_test)
        })
        
    results_df = pd.DataFrame(results_df)
    
    print('\nFinished the hyperparameter tuning for reconstruction. Saving the results.')
    # Save output
    np.save(processed_data_path + "hyperparameters/latent_benchmark_final.npy", latents)
    np.save(processed_data_path + "hyperparameters/con_recon_benchmark_final.npy", con_recons)
    np.save(processed_data_path + "hyperparameters/cat_recon_benchmark_final.npy", cat_recons)
    np.save(processed_data_path + "hyperparameters/recon_acc_benchmark_final.npy", recon_acc)
    np.save(processed_data_path + "hyperparameters/loss_benchmark_final.npy", loss_train)
    np.save(processed_data_path + "hyperparameters/likelihood_benchmark_final.npy", likelihoods)
    np.save(processed_data_path + "hyperparameters/best_epochs_benchmark_final.npy", best_epochs)

    np.save(processed_data_path + "hyperparameters/test_latent_benchmark_final.npy", latents_tests)
    np.save(processed_data_path + "hyperparameters/test_con_recon_benchmark_final.npy", con_recons_tests)
    np.save(processed_data_path + "hyperparameters/test_cat_recon_benchmark_final.npy", cat_recons_tests)
    np.save(processed_data_path + "hyperparameters/test_recon_acc_benchmark_final.npy", recon_acc_tests)
    np.save(processed_data_path + "hyperparameters/test_loss_benchmark_final.npy", loss_tests)
    np.save(processed_data_path + "hyperparameters/test_likelihood_benchmark_final.npy", likelihood_tests)
    
    print('The results saved.\n')
    
    return(likelihood_tests, recon_acc_tests, recon_acc, results_df)

    
    
def optimize_stability(nHiddens, nLatents, nDropout, nBeta, repeat, nepochs, nLayers, batch_sizes, lrate, kldsteps, batchsteps, cuda, path, con_list, cat_list, continuous_weights, categorical_weights, seed):
    
    models, latents, embeddings, con_recons, cat_recons, recon_acc, los, likelihood = initiate_default_dicts(1, 7)
    
    print('Beginning the hyperparameter tuning for stability.\n')
    iters = itertools.product(nHiddens, nLatents, nLayers, nDropout, nBeta, range(repeat))
    for nHidden, nLatent, nl, drop, b, r in iters:
        combi = str([nHidden] * nl) + "+" + str(nLatent) + ", do: " + str(drop) +", b: " + str(b)
        print(combi)

        best_model, loss, ce, sse, KLD, train_loader, mask, kld_w, cat_shapes, con_shapes, best_epoch = train_model(cat_list, con_list, categorical_weights, continuous_weights, batch_sizes, nHidden, nl, nLatent, b, drop, cuda, kldsteps, batchsteps, nepochs, lrate, seed, test_loader=None, patience=None, early_stopping=False)


        test_loader = DataLoader(dataset=train_loader.dataset, batch_size=1, drop_last=False,
                     shuffle=False, pin_memory=train_loader.pin_memory) #num_workers=1,

        latent, latent_var, cat_recon, cat_class, con_recon, test_loss, test_likelihood = best_model.latent(test_loader, kld_w)
        con_recon = np.array(con_recon)
        con_recon = torch.from_numpy(con_recon)

        cat_true_recon = cal_cat_recon(cat_shapes, cat_recon, cat_class)
        true_recon = cal_con_recon(train_loader, con_recon, con_shapes)

        ### Umap clustering 
        reducer = umap.UMAP()
        embedding = reducer.fit_transform(latent)

        # save
        recon_acc[combi].append(cat_true_recon + true_recon)
        latents[combi].append(latent)
        embeddings[combi].append(embedding)
        con_recons[combi].append(con_recon)
        cat_recons[combi].append(cat_recon)
        

    # Saving the results
    print('\nFinished the hyperparameter tuning for stability. Saving the results.')
    
    np.save(path + "hyperparameters/embedding_stab.npy", embeddings)
    np.save(path + "hyperparameters/latent_stab.npy", latents)
    np.save(path + "hyperparameters/con_recon_stab.npy", con_recons)
    np.save(path + "hyperparameters/cat_recon_stab.npy", cat_recons)
    np.save(path + "hyperparameters/recon_acc_stab.npy", recon_acc)
    
    print('The results saved.\n')
    
    return(embeddings, latents, con_recons, cat_recons, recon_acc)


def train_model_association(path, cuda, nepochs, nLatents, batch_sizes, nHidden, nl, nBeta, drop, con_list, cat_list, continuous_weights, categorical_weights, version, repeats, kldsteps, batchsteps, lrate, drug, categorical_names, data_of_interest, seed):

    results, recon_results, recon_results_1, mean_bas = initiate_default_dicts(n_empty_dicts=0, n_list_dicts=4)

    start, end = get_start_end_positions(cat_list, categorical_names, data_of_interest)


    iters = itertools.product(nLatents, range(repeats))

    # Running the framework    
    for nLatent, repeat in iters: 
        best_model, loss, ce, sse, KLD, train_loader, mask, kld_w, cat_shapes, con_shapes, best_epoch = train_model(cat_list, con_list, categorical_weights, continuous_weights, batch_sizes, nHidden, nl, nLatent, nBeta, drop, cuda, kldsteps, batchsteps, nepochs, lrate, seed, test_loader=None, patience=None, early_stopping=False)


        train_test_loader = DataLoader(dataset=train_loader.dataset, batch_size=train_loader.batch_size, drop_last=False,
                              shuffle=False, pin_memory=train_loader.pin_memory) #num_workers=1,

        latent, latent_var, cat_recon, cat_class, con_recon, loss, likelihood = best_model.latent(train_test_loader, kld_w)

        con_recon = np.array(con_recon)
        con_recon = torch.from_numpy(con_recon)

        mean_baseline = get_baseline(best_model, train_test_loader, con_recon, repeat=1, kld_w=kld_w)
        recon_diff, groups = change_drug(best_model, train_test_loader, con_recon, drug, start, end, kld_w) #TODO: when the drug data starts 
        stat = cal_sig_hits(recon_diff, groups, drug, mean_baseline, train_loader.dataset.con_all)

        results[nLatent].append(stat)
        recon_diff_corr = dict()
        for r_diff in recon_diff.keys():
            recon_diff_corr[r_diff] = recon_diff[r_diff] - np.abs(mean_baseline[groups[r_diff]])

        mean_bas[nLatent].append(mean_baseline)
        recon_results[nLatent].append(recon_diff_corr)
        recon_results_1[nLatent].append(recon_diff)


    # Saving results
    isExist = os.path.exists(path + 'wp2.2/sig_overlap/')
    if not isExist:
        os.makedirs(path + 'wp2.2/sig_overlap/')

    isExist = os.path.exists(path + 'results/')
    if not isExist:
        os.makedirs(path + 'results/')

    with open(path + "results/results_" + version + ".npy", 'wb') as f:
        np.save(f, results)
    with open(path + "results/results_recon_" + version + ".npy", 'wb') as f:
        np.save(f, recon_results)
    with open(path + "results/results_groups_" + version + ".npy", 'wb') as f:
        np.save(f, groups)
    with open(path + "results/results_recon_mean_baseline_" + version + ".npy", 'wb') as f:
        np.save(f, mean_bas)
    with open(path + "results/results_recon_no_corr_" + version + ".npy", 'wb') as f:
        np.save(f, recon_results_1)

    cor_results = correction_new(results)

    with open(path + "wp2.2/sig_overlap/cor_results_" + version + ".npy", 'wb') as f:
        np.save(f, cor_results)
        
    
    
def train_model(cat_list, con_list, categorical_weights, continuous_weights, batch_size, nHidden, nl, nLatent, b, drop, cuda, kldsteps, batchsteps, nepochs, lrate, seed, test_loader, patience, early_stopping):
    
    device = torch.device("cuda" if cuda == True else "cpu")
    
    if seed is not None:
        set_global_seed(seed)
    
    # Initiate loader
    mask, train_loader = dataloaders.make_dataloader(cat_list=cat_list, 
                                                     con_list=con_list, 
                                                     batchsize=batch_size)


    #          ncategorical = train_loader.dataset.cat_all.shape[1]
    #          ncontinuous = train_loader.dataset.con_all.shape[1]
    con_shapes = train_loader.dataset.con_shapes
    cat_shapes = train_loader.dataset.cat_shapes

    model = vae.VAE(continuous_shapes=con_shapes, 
                 categorical_shapes=cat_shapes, 
                 num_hidden=[nHidden]*nl,
                 num_latent=nLatent,  
                 beta=b, 
                 continuous_weights=continuous_weights,
                 categorical_weights=categorical_weights, 
                 dropout=drop, 
                 cuda=cuda).to(device) # removed alpha=0.01,       

    # Create lists for saving loss
    loss = list();ce = list();sse = list();KLD = list();loss_test = list()


    l = len(kldsteps) # TODOs: not really clear this l
    rate = 20/l 
    kld_w = 0
    update = 1 + rate
    l_min = float('inf')
    count = 0

    # Train model
    for epoch in range(1, nepochs + 1):
        if epoch in kldsteps:
            kld_w = 1/20 * update
            #lrate = lrate - lrate*0.1
            update += rate

        if epoch in batchsteps:
            train_loader = DataLoader(dataset=train_loader.dataset,batch_size=int(train_loader.batch_size * 1.5),shuffle=True,drop_last=True) #,num_workers=train_loader.num_workers,pin_memory=train_loader.pin_memory  Is it really?: batch_size=int(train_loader.batch_size * 1.5)

        l, c, s, k = model.encoding(train_loader, epoch, lrate, kld_w)


        loss.append(l)
        ce.append(c)
        sse.append(s)
        KLD.append(k)

        if early_stopping:

            out = model.latent(test_loader, kld_w)
            loss_test.append(out[-2])
            print("Likelihood: " + str(out[-1]))


            if out[-1] > l_min and count < patience: # Change to infinite value
                count+=1
                
                if count % 5 == 0:
                    lrate = lrate - lrate*0.10


            elif count == patience: #Changed from 100
                break

            else:
                l_min = out[-1]
                best_model = copy.deepcopy(model)
                best_epoch = epoch
                count = 0

#                 if epoch > 3: # Added
#                     break      # Added

#             if epoch > 10: # Added
#                 break      # Added    
        else:
            best_model = copy.deepcopy(model)
            best_epoch = None
        #                loss_test = None 

    return(best_model, loss, ce, sse, KLD, train_loader, mask, kld_w, cat_shapes, con_shapes, best_epoch)        


import os
import torch
import numpy as np

from torch.utils.data import DataLoader
from scipy.stats.stats import pearsonr
from sklearn.metrics.pairwise import cosine_similarity
from collections import defaultdict
import itertools 
import tqdm    
from scipy import stats
from statsmodels.stats.multitest import multipletests

from move.utils import dataloaders
# Funtions for calculations

def get_latent(best_model, train_loader, test_loader, kld_w):
    # Get training set results
    train_test_loader = DataLoader(dataset=train_loader.dataset, batch_size=1, drop_last=False, shuffle=False) #, num_workers=1, pin_memory=train_loader.pin_memory

    latent, latent_var, cat_recon, cat_class, con_recon, loss, likelihood = best_model.latent(train_test_loader, kld_w)
    con_recon = np.array(con_recon)

    con_recon = torch.from_numpy(con_recon)
    
    # Get test set results
    test_loader = DataLoader(dataset=test_loader.dataset, batch_size=1, drop_last=False, shuffle=False) #, num_workers=1, pin_memory=test_loader.pin_memory
    latent_test, latent_var_test, cat_recon_test, cat_class_test, con_recon_test, loss_test, likelihood_test = best_model.latent(test_loader, kld_w)
    con_recon_test = np.array(con_recon_test)
    con_recon_test = torch.from_numpy(con_recon_test)
    
    return con_recon, train_test_loader, latent, latent_var, cat_recon, cat_class, loss, likelihood, latent_test, latent_var_test, cat_recon_test, cat_class_test, con_recon_test, loss_test, likelihood_test

def cal_recon(cat_shapes, cat_recon, cat_class, train_loader, con_recon, con_shapes, cat_recon_test, cat_class_test, test_loader, con_recon_test):
    cat_true_recon = cal_cat_recon(cat_shapes, cat_recon, cat_class)
    true_recon = cal_con_recon(train_loader, con_recon, con_shapes)
    cat_true_recon_test = cal_cat_recon(cat_shapes, cat_recon_test, cat_class_test)
    true_recon_test = cal_con_recon(test_loader, con_recon_test, con_shapes)
    
    return cat_true_recon,true_recon,cat_true_recon_test,true_recon_test


# def save_input(combi,cat_true_recon,true_recon,latent,con_recon,cat_recon,loss,likelihood,best_epoch,cat_true_recon_test,true_recon_test,latent_test,con_recon_test,cat_recon_test,loss_test,likelihood_test):
  

#    recon_acc[combi].append(cat_true_recon + true_recon)
#    latents[combi].append(latent)
#    con_recons[combi].append(con_recon)
#    cat_recons[combi].append(cat_recon)
#    loss_train[combi].append(loss)
#    likelihoods[combi].append(likelihood)
#    best_epochs[combi].append(best_epoch)
  
#    recon_acc_tests[combi].append(cat_true_recon_test + true_recon_test)
#    latents_tests[combi].append(latent_test)
#    con_recons_tests[combi].append(con_recon_test)
#    cat_recons_tests[combi].append(cat_recon_test)
#    loss_tests[combi].append(loss_test)
#    likelihood_tests[combi].append(likelihood_test)
#    return(recon_acc, latents, con_recons, cat_recons, loss_train, likelihoods, best_epochs, recon_acc_tests,\
#             latents_tests, con_recons_tests, cat_recons_tests, loss_tests, likelihood_tests)

def cal_cat_recon(cat_shapes, cat_recon, cat_class):
    cat_true_recon = []
    cat_total_recon = []
    pos = 0
    for s in cat_shapes:
        n = s[1]
        cat_class_tmp = cat_class[:,pos:(n + pos)]
        cat_recon_tmp = cat_recon[:,pos:(n + pos)]
        
        missing_cat = cat_recon_tmp[cat_class_tmp == -1]
        diff_cat = cat_class_tmp - cat_recon_tmp
        
        diff_cat[diff_cat != 0] = -1
        true_cat = diff_cat[diff_cat == 0]
        false_cat = diff_cat[diff_cat != 0]
        cat_true = len(true_cat)/(float(diff_cat.size) - missing_cat.size)
        cat_true_recon.append(cat_true)
        diff_cat[diff_cat == 0] = 1
        diff_cat[diff_cat != 1] = 0
        cat_total_recon.append(np.count_nonzero(diff_cat, 1) / diff_cat.shape[1])
        pos += n
    return cat_true_recon

def cal_con_recon(train_loader, con_recon, con_shapes):
    total_shape = 0
    true_recon = []
    cos_values = []
    all_values = []
    for s in con_shapes:
        cor_con = list()
        cos_con = list()
        all_val =list()
        for n in range(0, con_recon.shape[0]):
            con_no_missing = train_loader.dataset.con_all[n,total_shape:(s + total_shape - 1)][train_loader.dataset.con_all[n,total_shape:(s + total_shape - 1)] != 0]
            if len(con_no_missing) <= 1:
                all_val.append(np.nan)
                continue
            con_out_no_missing = con_recon[n,total_shape:(s + total_shape - 1)][train_loader.dataset.con_all[n,total_shape:(s + total_shape - 1)] != 0]
            


            cor = pearsonr(con_no_missing, con_out_no_missing)[0]
            cor_con.append(cor)
            
            com = np.vstack([con_no_missing, con_out_no_missing])
            cos = cosine_similarity(com)[0,1]
            cos_con.append(cos)
            all_val.append(cos)
        
        cor_con = np.array(cor_con)
        cos_con = np.array(cos_con)
        cos_values.append(cos_con)
        all_values.append(np.array(all_val))
        true_recon.append(len(cos_con[cos_con >= 0.7]) / len(cos_con))
        total_shape += s
        
    return true_recon

# Functions for calculations

def get_baseline(model, train_loader, con_recon, repeat=25, kld_w=0):
    recon_diff_baseline = list()
    for r in range(repeat):
        latent_new, latent_var_new, cat_recon_new, cat_class_new, con_recon_new, test_loss_new, likelihood_new = model.latent(train_loader, kld_w)
    
        recon_diff = con_recon_new - np.array(con_recon) 
        recon_diff_baseline.append(recon_diff)
    
    matrix = np.array(recon_diff_baseline)
    baseline_mean = np.mean(matrix, axis= 0)
    baseline_mean = np.where(train_loader.dataset.con_all == 0, np.NaN , baseline_mean)
    return baseline_mean

def change_drug_atc(train_loader, trans_table, con_recon, drug, types = [[1,0]], data_start=1557):
    types = np.array(types)
    
    recon_diff_con_none = dict()
    none_groups = dict()
    data_shapes = drug.shape
    data_end = train_loader.dataset.cat_all.shape[1]
    for feature_index in range(data_shapes[1]):
        atc = trans_table[drug_h[feature_index]]
        same_atc = [k for k, v in trans_table.items() if (v == atc and k != drug_h[feature_index])]
        same = 0
        drug_indexes = []
        if len(same_atc) > 0:
            same = 1
            for i in same_atc:
                drug_indexes.append(drug_h.index(i))
        
        data = np.array(train_loader.dataset.cat_all)
        
        for t in types:
            tmp_data = np.copy(data[:,data_start:data_end])
            tmp_data = tmp_data.reshape(tmp_data.shape[0], data_shapes[1], data_shapes[2])
            tmp_data[:,feature_index,:] = t
            
            same_group = [True] * data.shape[0]
            none_group = [True] * data.shape[0]
            if same != 0:
                for i in drug_indexes:
                    same_group_tmp = [(np.all(a_s == types[0])) for a_s in drug[:,i,:]]
                    none_group_tmp = [not (np.all(a_s == types[0]) or (np.all(a_s == [0,0]))) for a_s in drug[:,i,:]]
                    
                    same_group = [all(tup) for tup in zip(same_group, same_group_tmp)]
                    none_group = [all(tup) for tup in zip(none_group, none_group_tmp)]
            
            tmp_data = tmp_data.reshape(tmp_data.shape[0], -1)
            new_data = np.copy(data)
            new_data[:,data_start:data_end] = tmp_data
            new_data = torch.from_numpy(new_data)
            
            none_groups[feature_index] = none_group
            new_data = new_data[none_group,:]
            dataset = dataloaders.MOVEDataset(new_data, train_loader.dataset.con_all[none_group,:], train_loader.dataset.con_shapes, train_loader.dataset.cat_shapes)
            
            new_loader = DataLoader(dataset, batch_size=train_loader.batch_size, drop_last=False,
                                    shuffle=False, num_workers=1, pin_memory=train_loader.pin_memory)
            
            latent_new, latent_var_new, cat_recon_new, cat_class_new, con_recon_new, test_loss_new, likelihood_new = model.latent(new_loader, kld_w)
            
            recon_diff = (con_recon_new - np.array(con_recon[none_group,:]))
            recon_diff_con_none[feature_index] = recon_diff
    
    return recon_diff_con_none,none_groups


def change_drug(model, train_loader, con_recon, drug, start, end, kld_w, types = [[1,0]]):
    types = np.array(types)
    
    recon_diff_con_none = dict()
    none_groups = dict()
    data_shapes = drug.shape
    data_start = start
    data_end = end #train_loader.dataset.cat_all.shape[1] # todo: Change to the end of the category 
    for feature_index in range(data_shapes[1]):
        data = np.array(train_loader.dataset.cat_all)
        for t in types:
            tmp_data = np.copy(data[:,data_start:data_end])
            tmp_data = tmp_data.reshape(tmp_data.shape[0], data_shapes[1], data_shapes[2])
            tmp_data[:,feature_index,:] = t
            
            none_group = [True] * data.shape[0]
            
            tmp_data = tmp_data.reshape(tmp_data.shape[0], -1)
            new_data = np.copy(data)
            new_data[:,data_start:data_end] = tmp_data
            new_data = torch.from_numpy(new_data)
            
            none_groups[feature_index] = none_group
            dataset = dataloaders.MOVEDataset(new_data, train_loader.dataset.con_all, train_loader.dataset.con_shapes, train_loader.dataset.cat_shapes)
            
            new_loader = DataLoader(dataset, batch_size=train_loader.batch_size, drop_last=False,
                                    shuffle=False,  pin_memory=train_loader.pin_memory) #num_workers=1,
            
            latent_new, latent_var_new, cat_recon_new, cat_class_new, con_recon_new, test_loss_new, likelihood_new = model.latent(new_loader, kld_w)
            
            recon_diff = (con_recon_new - np.array(con_recon))
            recon_diff_con_none[feature_index] = recon_diff
    
    return recon_diff_con_none,none_groups


def cal_sig_hits(recon_diff_con_none, none_groups, drug, baseline_mean, con_all, types = [[1,0]]):
    data_shapes = drug.shape
    none_avg = list()
    none_stats = list()
    for f in range(0,data_shapes[1]):
        if f in none_groups:
            tmp = np.copy(recon_diff_con_none[f])
            tmp_data = con_all[none_groups[f],:]
            tmp_baseline = baseline_mean[none_groups[f],:]
            g = [not (np.all(a_s == types[0]) or (np.all(a_s == [0,0]))) for a_s in drug[none_groups[f],f,:]]
            tmp = np.where(tmp_data[g,:] == 0,np.NaN, tmp[g,:])
            tmp_abs = np.abs(tmp)
            stat = stats.ttest_rel(tmp, tmp_baseline[g,:], axis=0, nan_policy="omit")
            
            if np.all(np.isnan(stat[1])):
                p_stat = np.zeros((con_all.shape[1]))
                p_stat[:] = np.nan
            else:
                p_stat = stat[1]
            
            none_stats.append(p_stat)
            avg = np.nanmean(tmp, axis = 0)
            avg[np.isnan(avg)] = 0
            none_avg.append(avg)
        else:
            tmp = np.zeros((tmp_data.shape[1]))
            tmp[:] = np.nan
            none_avg.append(tmp)
            none_stats.append(tmp)
    
    none_avg = np.array(none_avg)
    none_stats = np.array(none_stats)
    
    return none_stats

def correction_new(results):
    new_results = defaultdict(dict)
    for l in results:
        for r in range(len(results[l])):
            corrected = np.zeros((results[l][r].shape[0],results[l][r].shape[1]))
            for d in range(results[l][r].shape[0]):
                stats_cor = multipletests(results[l][r][d,:], method = "bonferroni")[1]
                corrected[d,:] = stats_cor
            
            new_results[r][l] = corrected
    
    return new_results

# def get_start_end_positions(cat_list, data_dict):
  
#    n_cat = 0
#    cat_shapes = list()
#    cat_all = []
#    first = 0
#    for i in range(len(data_dict['categorical_data_files'])):
#      cat_d = cat_list[i]
     
#      cat_shapes.append(cat_d.shape)
#      cat_input = cat_d.reshape(cat_d.shape[0], -1)
     
     
#      if first == 0:
#         cat_all = cat_input
#         del cat_input
#         first = 1
     
#         if data_dict['data_of_interest'] == data_dict['categorical_data_files'][i]:
#            start = 0
#            end = cat_all.shape[-1]
          
#      else:
#         cat_all = np.concatenate((cat_all, cat_input), axis=1)
        
#         if data_dict['data_of_interest'] == data_dict['categorical_data_files'][i]:
#            start = cat_all.shape[-1] - cat_input.shape[-1]
#            end = cat_all.shape[-1]
     
#    # Make mask for patients with no measurments
#    catsum = cat_all.sum(axis=1)
#    mask = catsum > 5
#    del catsum
#    return start, end



def get_start_end_positions(cat_list, categorical_names, data_of_interest):
  
    n_cat = 0
    cat_shapes = list()
    cat_all = []
    first = 0
    for i in range(len(categorical_names)):
        cat_d = cat_list[i]
     
        cat_shapes.append(cat_d.shape)
        cat_input = cat_d.reshape(cat_d.shape[0], -1)
     
     
        if first == 0:
            cat_all = cat_input
            del cat_input
            first = 1
     
            if data_of_interest == categorical_names[i]:
                start = 0
                end = cat_all.shape[-1]
          
        else: 
            cat_all = np.concatenate((cat_all, cat_input), axis=1)
        
            if data_of_interest == categorical_names[i]:
                start = cat_all.shape[-1] - cat_input.shape[-1]
                end = cat_all.shape[-1]
     
    # Make mask for patients with no measurments
    catsum = cat_all.sum(axis=1)
    mask = catsum > 5
    del catsum
    return start, end



import os 
import numpy as np
from collections import defaultdict
from omegaconf import OmegaConf

# from omegaconf import OmegaConf
# def merge_configs(base_config, config_type):
#     """Composes configuration for the MOVE framework.

#     Parameters
#     ----------
#     filepath : Union[str, Path]
#         Path to YAML configuration file

#     Returns
#     -------
#     move.conf.MOVEConfig
#     """
# #     print(OmegaConf.to_yaml(config))
# #     base_config = getattr(base_config, config_type)
#     user_config = OmegaConf.load(base_config.data.user_config)

#     config = OmegaConf.merge(base_config, user_config)
    
#     print(OmegaConf.to_yaml(config))
#     return config

def merge_configs(base_config, config_types):
    """Composes configuration for the MOVE framework.

    Parameters
    ----------
    filepath : Union[str, Path]
        Path to YAML configuration file

    Returns
    -------
    move.conf.MOVEConfig
    """
    user_config_dict = dict()
    for config_type in config_types:
        # Getting name of user config file and loading it 
        user_config_name = base_config[config_type]['user_config']
        user_config = OmegaConf.load(user_config_name)

        # Making dict with the same key as in configuration file
        user_config_dict[config_type] = user_config

    # Merging the base and user defined config file
    config = OmegaConf.merge(base_config, user_config_dict)

    # Getting a subsection of data used for printing 
    config_section_dict = {x: config[x] for x in config_types if x in config}
    config_section = OmegaConf.create(config_section_dict)

    print(f'\nConfiguration used: \n---\n{OmegaConf.to_yaml(config_section)}---\n')
    return(config_section)


# Functions for loading data
def read_cat(file):
    data = np.load(file)
    data = data.astype(np.float32)
    data_input = data.reshape(data.shape[0], -1)
    
    return data, data_input

def read_con(file):
    data = np.load(file)
    data = data.astype(np.float32)
    data[np.isnan(data)] = 0
    consum = np.absolute(data).sum(axis=0)
    mask_col = consum != 0
    data = data[:,mask_col]
    return data, mask_col

def read_header(file, mask=None, start=1):
    with open(file, "r") as f:
        h = f.readline().rstrip().split("\t")[start:]
    if not mask is None:
        h = np.array(h)
        h = h[mask]
    
    return h

def initiate_default_dicts(n_empty_dicts, n_list_dicts):
    
    dicts = [defaultdict() for _ in range(n_empty_dicts)] + \
              [defaultdict(list) for _ in range(n_list_dicts)]
     
    return(tuple(dicts))

def get_data(raw_data_path, interim_data_path, categorical_data, continuous_data, data_of_interest):
     
    # Initiate lists
    cat_list, cat_names, con_list, con_names = [], [], [], []
     
    # Get categorical variables
    for cat_data in categorical_data:
        cat, cat_input = read_cat(interim_data_path + f"{cat_data}.npy")
        cat_h = read_header(raw_data_path + f"{cat_data}.tsv")
          
        cat_list.append(cat)
        cat_names.append(cat_h)

     # Get continuous variables
    for con_data in continuous_data:
        con, con_mask = read_con(interim_data_path + f"{con_data}.npy")
        con_h = read_header(raw_data_path + f"{con_data}.tsv", con_mask)
          
        con_list.append(con)
        con_names.append(con_h)
 
    #Change data types
    headers_all = tuple(cat_names+con_names)
    con_names = np.concatenate(con_names)
    cat_names = np.concatenate(cat_names)

    # Select dataset of interest
    if data_of_interest in categorical_data:
        drug, drug_input = read_cat(interim_data_path + f"{data_of_interest}.npy")
        drug_h = read_header(raw_data_path + f"{data_of_interest}.tsv")
    elif data_of_interest in continuous_data:
        drug, drug_mask = read_con(interim_data_path + f"{data_of_interest}.npy")
        drug_h = read_header(raw_data_path + f"{data_of_interest}.tsv", drug_mask)  
    else:
        raise ValueError("""In data.yaml file data_of_interest is chosen neither
                                 from defined continuous nor categorical data types""")
     
    return(cat_list, con_list, cat_names, con_names, headers_all, drug, drug_h)


# Functions for encoding

def encode_cat(raw_input, na='NA'):
    """
    Encodes categorical data into one-hot encoding
    
    inputs:
         raw_input: a list of source data sorted by IDs from baseline_ids.txt file
    returns:
         data_input: one hot encoded data
    """ 
     
    matrix = np.array(raw_input)
    n_labels = matrix.shape[1]
    n_samples = matrix.shape[0]
    
    unique_sorted_data = np.unique(raw_input)
    num_classes = len(unique_sorted_data[~np.isnan(unique_sorted_data)])
    uniques = [*range(num_classes), 'nan']
 
    # make endocding dict
    encodings = defaultdict(dict)
    count = 0
    no_unique = 0
    
    for u in uniques:
        if u == na:
            encodings[u] = np.zeros(num_classes)
            continue
        encodings[u] = np.zeros(num_classes)
        encodings[u][count] = 1
        count += 1

    # encode the data
    data_input = np.zeros((n_samples,n_labels,num_classes))
    i = 0
    for patient in matrix:
        
        data_sparse = np.zeros((n_labels, num_classes))
        count = 0
        for lab in patient:
            if no_unique == 1:
                data_sparse[count] = encodings[count][lab]
            else:
                if lab != na:
                    lab = int(float(lab))
                data_sparse[count] = encodings[lab]
            count += 1
        
        data_input[i] = data_sparse
        i += 1
        
    return data_input

def encode_con(raw_input):
    """
    Log transforms and z-normalizes the data
    
    Input: 
         raw_input: a list of source data sorted by IDs from baseline_ids.txt file
    Returns:
         data_input: numpy array with log transformed and z-score normalized data
         mask_col: a np.array vector of Bolean values that correspond to nonzero sd values 
    """

    matrix = np.array(raw_input)
    consum = matrix.sum(axis=1)
    
    data_input = np.log2(matrix + 1) 
    
    # remove 0 variance
    std = np.nanstd(data_input, axis=0)
    mask_col = std != 0
    data_input = data_input[:,mask_col]
     
    # z-score normalize
    mean = np.nanmean(data_input, axis=0)
    std = np.nanstd(data_input, axis=0)
    data_input -= mean
    data_input /= std
    return data_input, mask_col 


def sort_data(data, ids, labels):
    """
    Sorts data based on the ids file
    
    Inputs:
         data: a dictionary with the data to encode
         ids: a list of personal identfiers (ID) from baseline_ids.txt file
         labels: a list of column names from the source data file
    Returns:
         sorted_data: a list of source data sorted by IDs from baseline_ids.txt file
    """

    n_labels = len(labels)
    sorted_data = list()

    for _ids in ids: #check: ids/ids
        if _ids in data:
            sorted_data.append(data[_ids])
        else:
            tmp = [0]*n_labels
            sorted_data.append(tmp)
    return sorted_data

def read_files(path, data_type, ids_file_name, na):
    """
    Function reads the input file into the dictionary
     
    Inputs:
         data_type: a string that defines a name of .tsv file to encode
         na: a string that defines how NA values are defined in the source data file
    Returns:
         ids: a list of personal identfiers (ID) from baseline_ids.txt file
         raw_input: a dictionary with the data to encode
         header: a list of column names from the source data file
    """
     
    ids = list()
    with open(path + f"{ids_file_name}.txt", "r") as f:
        for line in f:
            ids.append(line.rstrip()) 
                 
    raw_input = dict()
    with open(path + f"{data_type}.tsv", "r") as f:
        header = f.readline()
        for line in f:
            line = line.rstrip()
            tmp = np.array(line.split("\t"))
            vals = tmp[1:]
            vals[vals == na] = np.nan
            vals = list(map(float, vals))
            raw_input[tmp[0]] = vals
    header = header.split("\t")
     
    return ids, raw_input, header

def generate_file(var_type, raw_data_path, interim_data_path, data_type, ids_file_name, na='NA'):
    """
    Function encodes source data type and saves the file
     
    inputs:
         var_type: a string out of ['categorical', 'continuous'], defines input data type to encode
         path: a string that defines a path to the directory the input data is stored
         data_type: a string that defines a name of .tsv file to encode
         na: a string that defines how NA values are defined in the source data file
    """
    
    # Preparing the data
    isExist = os.path.exists(interim_data_path)
    if not isExist:
        os.makedirs(interim_data_path)
    
    ids, raw_input, header = read_files(raw_data_path, data_type, ids_file_name, na)
    sorted_data = sort_data(raw_input, ids, header)
     
    if var_type == 'categorical':
        data_input = encode_cat(sorted_data, 'nan')
    elif var_type == 'continuous':
        data_input, _ = encode_con(sorted_data)
            
    np.save(interim_data_path + f"{data_type}.npy", data_input)

def get_list_value(*args):
    arg_tuple = [arg[0] if len(arg) == 1 else arg for arg in args]
    return(arg_tuple)     



import numpy as np
import matplotlib.pyplot as plt
plt.style.use('seaborn-whitegrid')

from sklearn.metrics.pairwise import cosine_similarity
import pandas as pd
import seaborn as sns
import matplotlib

from move.utils import plot_importance

def visualize_likelihood(path, nLayers, nHiddens, nDropout, nBeta, nLatents, likelihood_tests):
    # Figure for test error/likelihood
    
    # Define styles for the plot
    ncols = ['navy', 'forestgreen', 'dodgerblue']
    styles = [':', '-', '--']
    batch_size = 10 # here I only tested one batch size due to my small sample size
                    # This can be exhanged with another parameter
    n_rows = len(nLayers)
    n_cols = len(nHiddens)

    
    fig, axes = plt.subplots(n_rows, n_cols, figsize=(18,15), sharex=True, sharey=True, frameon=False)
    ax3 = fig.add_subplot(111, frameon=False)
    
    # Plotting
    y = 0
    for nl in nLayers:
        x = 0
        for nHidden in nHiddens:
            for drop in nDropout:
                c = ncols[nDropout.index(drop)]
                for b in nBeta:
                    s = styles[nBeta.index(b)]

                    d = []
                    for nLatent in nLatents:
                        combi = str([nHidden] * nl) + "+" + str(nLatent) + ", drop: " + str(drop) + \
                                   ", b: " + str(b) + ", batch: " + str(batch_size)
                        d.append(likelihood_tests[combi][0])

                    a = axes[x,y].plot(nLatents, d, label='Dropout: ' + str(drop) +
                                             ', beta: ' + str(b), linestyle=s, color = c, linewidth=2)
            x += 1
            
        y += 1
     
     
     ### Annotating the graph
    
    # Adding annotations of variables on right side of the subplots on y axis
    for i in range(n_rows):
        axes[i, n_cols-1].annotate(nHiddens[i], xy=(1.02, 0.55), xycoords='axes fraction',
                                fontsize = 14, rotation=-90)
    
    # Adding annotation of variable on right side of the graph
    plt.figtext(0.78, 0.40, 'Number of hidden neurons in each layer', rotation=-90, fontsize=14)
    
    # Adding annotations of variables on top of the subplots on x axis
    for i in range(n_cols):
        axes[0,i].set_title(nLayers[i], fontsize=14)
    
    # Setting the size of variables on ticks
    for i in range(n_rows):
        for j in range(n_cols):
            axes[i,j].xaxis.set_tick_params(labelsize=14)
            axes[i,j].yaxis.set_tick_params(labelsize=14)
    plt.xticks(nLatents, nLatents, fontsize=14)
    ax3.tick_params(labelcolor='none', top=False, bottom=False, left=False, right=False)
    
    # Setting labels on X and Y axis
    ax3.set_xlabel('Number of latent neurouns', fontsize = 14)
    ax3.set_ylabel('Log-likelihood', fontsize = 14)
    ax3.set_title('Number of hidden layers', fontsize = 14)
    ax3.title.set_position([.5, 1.03])
    ax3.yaxis.set_label_coords(-0.04, 0.5)
    ax3.xaxis.set_label_coords(0.5, -0.03)
    
    # Adding additional adjustments to the graph
    plt.grid(False)
    plt.tick_params(labelcolor='none', top=False, bottom=False, left=False, right=False)
    plt.tight_layout()
    fig.subplots_adjust(right=0.75)
    handles, labels = axes[0,0].get_legend_handles_labels()
    fig.legend(handles, labels, loc='center right', fontsize = 14)
    
    # Saving the figure
    plt.savefig(path + "hyperparameters/all_likelihoods_test1.png")

    
def visualize_recon_acc(path, nLayers, nHiddens, nDropout, nBeta, nLatents, recon_acc, data_type):
    # Plot results for train reconstructions
    n_rows = len(nLayers)
    n_cols = len(nHiddens)
    
    fig, axes = plt.subplots(n_rows, n_cols, figsize=(40,30), sharex=True, sharey=True, frameon=False)
    ax3 = fig.add_subplot(111, frameon=False)

    y = 0
    nBeta = [nBeta[0]] # Beta didn't really effect the results so for the reconstructions we only looked at one value
    batch_size = 10
    # Plotting
    for nl in nLayers:
        x = 0
        for nHidden in nHiddens:
            frame = pd.DataFrame()
            indexes = []
            for drop in nDropout:
                for b in nBeta:
                    for nLatent in nLatents:
                        combi = str([nHidden] * nl) + "+" + str(nLatent) + ", drop: " + str(drop) + ", b: " + str(b) + ", batch: " + str(batch_size)
                        r = recon_acc[combi][0]
                    
                        name = 'Latent: ' + str(nLatent) + ', Dropout: ' + str(drop)
                        indexes.append(name)
                        frame[combi] = r

            frame.set_axis(indexes, axis=1, inplace=True)
            sns.boxplot(data=frame, palette="colorblind", width=0.7, ax = axes[x,y])    
          
            x += 1
        y += 1

    ### Anotating
    # Adding annotations of variables on right side of the subplots on y axis
    for i in range(n_rows):
        axes[i,n_cols-1].annotate(nHidden, xy=(1.02, 0.55), xycoords='axes fraction', fontsize = 24, rotation=-90)

    # Adding annotation of variable on right side of the graph
    plt.figtext(0.92, 0.45, 'Number of hidden neurons in each layer', rotation=-90, fontsize=24)
    
    # Adding annotations of variables on top of the subplots on x axis
    for i in range(n_cols):
        axes[0,i].set_title(nLayers[i], fontsize=24)
    
    # Setting the size of variables on ticks
    for i in range(n_rows):
        for j in range(n_cols):
            axes[i,j].set_xticklabels(indexes,fontsize=24, rotation=40, ha="right")
            axes[i,j].yaxis.set_tick_params(labelsize=24)
    
    # Setting labels on X and Y axis
    ax3.tick_params(labelcolor='none', top=False, bottom=False, left=False, right=False)
    ax3.set_ylabel('Reconstruction accuracy', fontsize = 24)
    ax3.set_title('Number of hidden layers', fontsize = 24)
    
    # Adding additional adjustments to the graph
    ax3.title.set_position([.5, 1.03])
    ax3.yaxis.set_label_coords(-0.04, 0.5)
    ax3.xaxis.set_label_coords(0.5, -0.03)
    plt.grid(False)
    plt.tick_params(labelcolor='none', top=False, bottom=False, left=False, right=False)
    plt.show()
    
    # Saving a figure
    plt.savefig(path + f"hyperparameters/all_recon_{data_type}.png")

    
def plot_graphs():
    fig = plt.figure()
    plt.plot(epochs, loss, '-g', label='loss')
    plt.plot(epochs, ce, '-b', label='CE')
    plt.plot(epochs, sse, '-r', label='SSE')
    plt.plot(epochs, KLD, '-y', label='KLD')
    plt.legend()
    plt.savefig(path + "/evaluation/loss_" + combi +".png")
    plt.clf()

    fig = plt.figure()
    plt.plot(epochs, loss, '-b', label='loss')
    plt.plot(batchsteps, loss_test, '-r', label='test loss')
    plt.legend()
    plt.savefig(path + "/evaluation/test_loss_" + combi +".png")
    plt.clf()    
    
    
def draw_boxplot(path, df, title_text, y_label_text, save_fig_name):
    
    df = pd.DataFrame(df)
    fig = plt.figure(figsize=(18,14))
    ax = sns.boxplot(data=df, palette = sns.color_palette('colorblind', df.shape[1]))
    ax.set_xticklabels(ax.get_xticklabels(),rotation=45, size=16, horizontalalignment='right')
    plt.title(title_text, size=20)
    plt.ylabel(y_label_text, size=16)
    plt.xlabel('')
    plt.yticks(fontsize=16)
    plt.savefig(path + f"hyperparameters/{save_fig_name}.png")
    

    

def embedding_plot_discrete(embedding, _type, name, file, palette=None):
    fig = plt.figure(figsize=(12,8))
    if palette == None:
        palette = sns.color_palette('colorblind', len(np.unique(_type)))

    ax = sns.scatterplot(x=embedding[:,0], y=embedding[:,1], hue=_type,
                         palette = palette,
                         linewidth=0.1, alpha = 0.8, s=40, edgecolor = 'black')


    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * 0.80, box.height]) # resize position

    legend_format = {name: np.unique(_type)}
    leg = subtitle_legend(ax, legend_format=legend_format)

    plt.xlabel('Dim1')
    plt.ylabel('Dim2')
    plt.style.use('default')
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(True)
    ax.spines['bottom'].set_visible(True)

    plt.savefig(file)

    
    
def embedding_plot_float(embedding, type, name, file):
    fig, ax = plt.subplots(figsize=(12,8))
    points = ax.scatter(x=embedding[:,0], y=embedding[:,1], c=type, s=40, cmap="Spectral_r",
                       edgecolor = 'black', linewidth=0.1)

    cbar = fig.colorbar(points, fraction=0.03, pad=0.03)
    cbar.ax.set_title(name, rotation=0, fontsize = 16)

    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * 0.90, box.height])

    plt.xlabel('Dim1')
    plt.ylabel('Dim2')
    plt.style.use('default')
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(True)
    ax.spines['bottom'].set_visible(True)

    plt.savefig(file)


def subtitle_legend(ax, legend_format):
    new_handles = []
   
    handles, labels = ax.get_legend_handles_labels()
    label_dict = dict(zip(labels, handles))
   
   #Means 2 labels were the same
    if len(label_dict) != len(labels):
        raise ValueError("Can not have repeated levels in labels!")
     
    for subtitle, level_order in legend_format.items():
       #Roll a blank handle to add in the subtitle
        blank_handle = matplotlib.patches.Patch(visible=False, label=subtitle)
        new_handles.append(blank_handle)
       
        for level in level_order:
            handle = label_dict[str(level)]
            new_handles.append(handle)
   
   #Labels are populated from handle.get_label() when we only supply handles as an arg
    legend = ax.legend(handles=new_handles, bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.,
                   fontsize = 14)
   #Turn off DrawingArea visibility to left justify the text if it contains a subtitle
    for draw_area in legend.findobj(matplotlib.offsetbox.DrawingArea):
        for handle in draw_area.get_children():
            if handle.get_label() in legend_format:
                draw_area.set_visible(False)
   
    return legend

def visualize_training(path, losses, ce, sse, KLD, epochs):
    # Plot traing error
    fig = plt.figure()
    plt.plot(epochs, losses, '-g', label='loss')
    plt.plot(epochs, ce, '-b', label='CE')
    plt.plot(epochs, sse, '-r', label='SSE')
    plt.plot(epochs, KLD, '-y', label='KLD')
    plt.legend()
    plt.show()
    plt.savefig(path + "loss_test.png")
    
def plot_reconstruction_distribs(path, cat_total_recon, all_values):
    
    # Plot the reconstruction distributions
    df = pd.DataFrame(cat_total_recon + all_values, index = ['Clinical\n(categorical)', 'Genomics', 'Drug data', 'Clinical\n(continuous)', 'Diet +\n wearables','Proteomics','Targeted\nmetabolomics','Untargeted\nmetabolomics', 'Transcriptomics', 'Metagenomics'])
    df_t = df.T

    fig = plt.figure(figsize=(25,15))
    ax = sns.boxplot(data=df_t, palette="colorblind", width=0.7)
    ax.set_xticklabels(ax.get_xticklabels(),rotation=45)
    plt.ylabel('Reconstruction accuracy', fontsize=24)
    ax.xaxis.set_tick_params(labelsize=22)
    ax.yaxis.set_tick_params(labelsize=22)
    fig.subplots_adjust(bottom=0.2)
    plt.savefig(path + "reconstruction_accuracy.png")
    plt.show()
    plt.close("all")    
    
def get_feature_data(data_type, feature_of_interest, cat_list, #TODO: negative values goes down to -2, while positive only until 1
                     con_list, cat_names, con_names):
    
    if data_type=='categorical':
        cat_list_integer = [np.argmax(cat, axis=-1) for cat in cat_list]
        np_data_ints = np.concatenate(cat_list_integer, axis=-1)
        headers = cat_names
    elif data_type=='continuous':
        np_data_ints = np.concatenate(con_list, axis=-1)
        headers = con_names
    else:
        raise ValueError("Wrong data type was selected")
    
    feature_data = np_data_ints[:,list(headers).index(feature_of_interest)]
    
    return(feature_data, headers)
    

def visualize_embedding(path, feature_of_interest, embedding, mask, cat_list, 
                        con_list, cat_names, con_names):
    
    if feature_of_interest in cat_names:
        data_type = 'categorical'
    elif feature_of_interest in con_names:
        data_type = 'continuous'
    else:
        raise ValueError("feature_of_interest is not in cat_names or con_names")
    
    
    feature_data, headers = get_feature_data(data_type, feature_of_interest, 
                                             cat_list=cat_list, con_list=con_list, 
                                             cat_names=cat_names, con_names=con_names)
    
    
    if data_type == 'categorical':
        embedding_plot_discrete(embedding, feature_data[mask], 
                                f"{feature_of_interest} (yes/no)", 
                                path + f"results/umap_{feature_of_interest}.png")
        
    elif data_type =='continuous':
        embedding_plot_float(embedding, feature_data[mask], feature_of_interest,
                             path + f"results/umap_{feature_of_interest}.png")
        
        
def f_plot_importance(path, sum_diffs, features, feature_names, fig_name):
    fig = plt.figure(figsize = (20,20))
    plot_importance.summary_plot(sum_diffs, features=features, 
                                 feature_names=feature_names, max_display = 25, 
                                 show = False, size = 30)
    plt.savefig(path + f"results/{fig_name}.png")
    
    
def plot_categorical_importance(path, sum_diffs, cat_list, feature_names, fig_name): #TODO: It is strangely the same plots for each SNP
    
    # Converting from one hot to numerical variables
    cat_ints_list = []
    for cat in cat_list:
        cat_target = np.argmax(cat, 2)
        cat_target[np.sum(cat, 2) == 0] = -1
        cat_ints_list.append(cat_target)
    
    cat_target_all = np.concatenate(cat_ints_list, axis=1)
    sum_diffs = np.transpose(sum_diffs)
    
    f_plot_importance(path=path,
                      sum_diffs=sum_diffs,
                      features=cat_target_all,
                      feature_names=feature_names,
                      fig_name='importance_SHAP_cat')
    

    
def plot_continuous_importance(path, train_loader, sum_diffs, feature_names, fig_name):
    
    con_all = np.asarray(train_loader.dataset.con_all)
    sum_diffs = np.transpose(sum_diffs)

    f_plot_importance(path=path,
                      sum_diffs=sum_diffs,
                      features=con_all,
                      feature_names=feature_names,
                      fig_name='importance_SHAP_con')
    

def visualize_indi_var(df_indi_var, version, path):

    plt.style.use('seaborn-whitegrid')
    fig = plt.figure(figsize=(12,10))
    ax = sns.barplot(data=df_indi_var.T, palette="tab10", saturation=0.50)
    ax.set_xticklabels(ax.get_xticklabels(),rotation=60, 
                      ha="right", rotation_mode="anchor")
    plt.ylabel('Patient variance')
    fig.subplots_adjust(bottom=0.2)
    fig.tight_layout()
    plt.grid(False)
    plt.savefig(path + "results/drug_individual_variations_" + version + ".pdf", format = 'pdf', dpi = 800)
    # plt.clf()
    plt.show()   


def visualize_drug_similarity_across_all(recon_average_corr_new_all, drug_h, version, path):
    
    cos_sim = cosine_similarity(recon_average_corr_new_all)

    corr = pd.DataFrame(cos_sim, columns = drug_h, index = drug_h)
    cmap = sns.diverging_palette(220, 10, as_cmap=True)
    sns.set(font_scale=1.5)
    f, ax = plt.subplots(figsize=(10, 10))

    g = sns.clustermap(corr, cmap=cmap, center=0, xticklabels = True,
                      yticklabels = True,
                      linewidths=0, row_cluster=True, col_cluster=True, metric='correlation')

    plt.savefig(path + "results/heatmap_" + version + "_all.pdf", format = 'pdf', dpi = 800)




# Import functions
import hydra 
from move.conf.schema import MOVEConfig

from move.training.train import optimize_stability
from move.utils.data_utils import get_data, get_list_value, merge_configs
from move.utils.visualization_utils import draw_boxplot
from move.utils.analysis import get_top10_stability, calculate_latent, make_and_save_best_stability_params # get_best_stability_paramset, get_best_4_latent_spaces

@hydra.main(config_path="../conf", config_name="main")
def main(base_config: MOVEConfig): 
    
    # Merging the user defined data.yaml, model.yaml and tuning_stability.yaml 
    # with the base_config to override it.
    print('Overriding the default config with configs from data.yaml, model.yaml and tuning_stability.yaml')
    cfg = merge_configs(base_config=base_config, 
                        config_types=['data', 'model', 'tuning_stability'])
    
    #Getting the variables used in the notebook
    raw_data_path = cfg.data.raw_data_path
    interim_data_path = cfg.data.interim_data_path
    processed_data_path = cfg.data.processed_data_path 
    
    data_of_interest = cfg.data.data_of_interest
    categorical_names = cfg.data.categorical_names
    continuous_names = cfg.data.continuous_names
    categorical_weights = cfg.data.categorical_weights
    continuous_weights = cfg.data.continuous_weights
    
    seed = cfg.model.seed
    cuda = cfg.model.cuda
    lrate = cfg.model.lrate
    kld_steps = cfg.model.kld_steps
    batch_steps = cfg.model.batch_steps
    
    nHiddens = cfg.tuning_stability.num_hidden
    nLatents = cfg.tuning_stability.num_latent
    nLayers = cfg.tuning_stability.num_layers
    nDropout = cfg.tuning_stability.dropout
    nBeta = cfg.tuning_stability.beta
    batch_sizes = cfg.tuning_stability.batch_sizes
    repeat = cfg.tuning_stability.repeats
    nepochs = cfg.tuning_stability.tuned_num_epochs
    
    # Raising the error if more than 1 batch size is used 
    if len(batch_sizes)==1:
        batch_sizes = batch_sizes[0]
    elif len(batch_sizes)>1:
        raise('Currently the code is implemented to take take only one value for batch_size')
    
    #Getting the data
    cat_list, con_list, cat_names, con_names, headers_all, drug, drug_h = get_data(raw_data_path, interim_data_path, categorical_names, continuous_names, data_of_interest)
    
    #Performing hyperparameter tuning
    embeddings, latents, con_recons, cat_recons, recon_acc = optimize_stability(nHiddens, nLatents, 
                                                                                nDropout, nBeta, repeat,
                                                                                nepochs, nLayers,
                                                                                batch_sizes, lrate, 
                                                                                kld_steps, batch_steps, 
                                                                                cuda, processed_data_path, 
                                                                                con_list, cat_list,
                                                                                continuous_weights, categorical_weights,
                                                                                seed)
    
    # Getting stability results 
    stability_top10, stability_top10_df = get_top10_stability(nHiddens, nLatents, nDropout, nLayers, repeat, latents, batch_sizes, nBeta)
    
    stability_total, rand_index = calculate_latent(nHiddens, nLatents, nDropout, repeat, nLayers, nBeta, latents) # Todo add priting or smth
    
    # Plotting the results 
    try: 
        draw_boxplot(path=processed_data_path ,
                     df=stability_top10,
                     title_text='Difference across replicationes in cosine similarity of ten closest neighbours in first iteration',
                     y_label_text="Average change",
                     save_fig_name="stability_top10")

        draw_boxplot(path=processed_data_path ,
                     df=stability_total,
                     title_text='Difference across replicationes in cosine similarity compared to first iteration',
                     y_label_text="Average change",
                     save_fig_name="stability_all")

        draw_boxplot(path=processed_data_path,
                     df=rand_index,
                     title_text='Rand index across replicationes compared to first iteration',
                     y_label_text="Rand index",
                     save_fig_name="rand_index_all")
        print('Visualizing the hyperparameter tuning results\n')
        
    except:
        print('Could not visualize the results\n')
    
    # Getting best set of hyperparameters
    hyperparams_names = ['num_hidden', 'num_latent', 'num_layers', 'dropout', 'beta', 'batch_sizes']
    make_and_save_best_stability_params(stability_top10_df, hyperparams_names, nepochs)

    return()
    
if __name__ == "__main__":
    main()



# Load functions
import hydra 
from move.conf.schema import MOVEConfig

from move.training.train import train_model_association
from move.utils.data_utils import get_data, merge_configs
from move.utils.visualization_utils import visualize_indi_var, visualize_drug_similarity_across_all
from move.utils.analysis import cal_reconstruction_change, overlapping_hits, identify_high_supported_hits, report_values, get_change_in_reconstruction, write_omics_results, make_files, get_inter_drug_variation, get_drug_similar_each_omics

import numpy as np 

@hydra.main(config_path="../conf", config_name="main")
def main(base_config: MOVEConfig): 
    
    # Merging the user defined data.yaml, model.yaml and training_association.yaml 
    # with the base_config to override it.
    print('Overriding the default config with configs from data.yaml, model.yaml and training_association.yaml')
    cfg = merge_configs(base_config=base_config, 
                        config_types=['data', 'model', 'training_association'])
    
    #Getting the variables used in the notebook
    raw_data_path = cfg.data.raw_data_path
    interim_data_path = cfg.data.interim_data_path
    processed_data_path = cfg.data.processed_data_path 
    data_of_interest = cfg.data.data_of_interest
    version = cfg.data.version
    categorical_names = cfg.data.categorical_names
    continuous_names = cfg.data.continuous_names
    categorical_weights = cfg.data.categorical_weights
    continuous_weights = cfg.data.continuous_weights
    up_down_list = cfg.data.write_omics_results_notebook5
    
    seed = cfg.model.seed
    cuda = cfg.model.cuda
    lrate = cfg.model.lrate
    kld_steps = cfg.model.kld_steps
    batch_steps = cfg.model.batch_steps

    nHiddens = cfg.training_association.num_hidden
    nLatents = cfg.training_association.num_latent
    nLayers = cfg.training_association.num_layers
    nDropout = cfg.training_association.dropout
    nBeta = cfg.training_association.beta
    batch_sizes = cfg.training_association.batch_sizes
    nepochs = cfg.training_association.tuned_num_epochs
    repeats = cfg.training_association.repeats
    
    types = [[1, 0]]
    
    # Checking if all data types selected for visualization are in continuous_names
    for data_type in up_down_list:
        if data_type not in continuous_names:
            raise ValueError(f"{data_type} is not in the continuous_names list.")
    
    # Getting the data
    cat_list, con_list, cat_names, con_names, headers_all, drug, drug_h = get_data(raw_data_path, interim_data_path, categorical_names, continuous_names, data_of_interest)
    
    # Training the model
    print('Beginning training the model.\n')
    train_model_association(processed_data_path, cuda, nepochs, nLatents, batch_sizes, nHiddens, nLayers, nBeta, nDropout, con_list, cat_list, continuous_weights, categorical_weights, version, repeats, kld_steps, batch_steps, lrate, drug, categorical_names, data_of_interest, seed)
    print('\nFinished training the model.')
    
    # Loading the saved files by train_model_association() - for using the results without the need to rerun the function
    results = np.load(processed_data_path + "results/results_" + version + ".npy", allow_pickle=True).item()
    recon_results = np.load(processed_data_path + "results/results_recon_" + version + ".npy", allow_pickle=True).item()
    groups = np.load(processed_data_path + "results/results_groups_" + version + ".npy", allow_pickle=True).item()
    mean_bas = np.load(processed_data_path + "results/results_recon_mean_baseline_" + version + ".npy", allow_pickle=True).item()
    recon_results_1 = np.load(processed_data_path + "results/results_recon_no_corr_" + version + ".npy", allow_pickle=True).item()
    cor_results = np.load(processed_data_path + "wp2.2/sig_overlap/cor_results_" + version + ".npy", allow_pickle=True).item()
    
    ### Starting the analysis
    
    # Getting the reconstruction average results
    recon_average = cal_reconstruction_change(recon_results, repeats)
    
    # Getting overlapping hits
    sig_hits, median_p_val = overlapping_hits(nLatents, cor_results, repeats, con_names, drug)
    
    # Getting high supported hits
    all_hits, collected_overlap = identify_high_supported_hits(sig_hits, drug_h, version, processed_data_path)
    
    # Saving the pi values of results of overlapping_hits() and  identify_high_supported_hits() functions
    report_values(processed_data_path, sig_hits, median_p_val, drug_h, all_hits, con_names)
    
    # Calculating average change among different runs
    con_list_concat = np.concatenate(con_list, axis=-1)
    recon_average_corr_new_all, recon_average_corr_all_indi_new = get_change_in_reconstruction(recon_average, groups, drug, drug_h, con_names, collected_overlap, sig_hits, con_list_concat, version, processed_data_path, types)
    
    # Loading the results saved by get_change_in_reconstruction() - for using the results without the need to rerun the function
    recon_average_corr_new_all = np.load(processed_data_path + "results/results_confidence_recon_all_" + version + ".npy", allow_pickle=True)
    recon_average_corr_all_indi_new = np.load(processed_data_path + "results/results_confidence_recon_all_indi_" + version + ".npy", allow_pickle=True).item()
    
    # Writing all the hits for each drug and database separately. Also, writing what features were increased or decreased with the association with the drug 
    write_omics_results(processed_data_path, up_down_list, collected_overlap, recon_average_corr_new_all, headers_all, continuous_names, data_of_interest)
    
    # Saving the effect sizes (95 % interval) of results of get_change_in_reconstruction() functions
    make_files(collected_overlap, groups, con_list_concat, processed_data_path, recon_average_corr_all_indi_new, con_names, continuous_names, drug_h, drug, all_hits, types, version)
    
    # Getting inter drug variation 
    df_indi_var = get_inter_drug_variation(con_names, drug_h, recon_average_corr_all_indi_new, 
                                           groups, collected_overlap, drug, con_list_concat, processed_data_path, types)

    # Visualizing variation, heatmap of similarities within drugs across all data and specific for each omics
    visualize_indi_var(df_indi_var, version, processed_data_path)
    visualize_drug_similarity_across_all(recon_average_corr_new_all, drug_h, version, processed_data_path)
    get_drug_similar_each_omics(con_names, continuous_names, all_hits, recon_average_corr_new_all, drug_h, version, processed_data_path)
    
if __name__ == "__main__":
    main()